import rich_click as click

from truefoundry.cli.const import GROUP_CLS
from truefoundry.common.utils import is_internal_env_set
from truefoundry.deploy.cli.commands.k8s_exec_credential_command import (
    k8s_exec_credential_command,
)
from truefoundry.deploy.cli.commands.kubeconfig_command import kubeconfig_command

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="get", cls=GROUP_CLS)
def get_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Get TrueFoundry resources

    \b
    Supported resources:
    - Kubeconfig
    """
    pass


def get_get_command():
    get_command.add_command(kubeconfig_command)
    if is_internal_env_set():
        get_command.add_command(k8s_exec_credential_command)
    return get_command



from __future__ import annotations

from typing import Dict, Optional

import docker
from openai import OpenAI
from rich.console import Console

from truefoundry.autodeploy.agents.base import Agent
from truefoundry.autodeploy.agents.project_identifier import ProjectIdentifier
from truefoundry.autodeploy.tools import (
    Ask,
    DockerRun,
    RequestEvent,
    ResponseEvent,
    SendRequest,
)
from truefoundry.pydantic_v1 import Field


class Tester(Agent):
    description = """
Tester
"""
    system_prompt = """
Your goal is to test a docker image.
In case the image is running a service, you can send a request to the endpoint to verify everything is running fine.
If you find any port number in the logs. Re run the docker image by exposing that port correctly.

If the image is not running a service, try to identify whether there is any issue.
Your goal is not to fix the issue. Your goal is to create a very detailed justification and report whether
the testing was successful.
Always response with a function call.
Return response once you are done testing.
    """
    max_iter = 30

    class Request(RequestEvent):
        project_identity: ProjectIdentifier.Response
        image_tag: str
        command: str
        port_to_be_exposed: Optional[int] = None

    class Response(ResponseEvent):
        successful: bool = Field(..., description="is everything fine?")
        justification: str = Field(
            ...,
            description="""
Why was the testing a failure or successful?
        """,
        )
        logs: str

        def render(self, console: Console):
            console.rule("[bold green]Result[/]")
            console.print(
                f"[bold cyan]TrueFoundry:[/] The given project has been {'[bold green]successfully built[/]' if self.successful else '[bold red]failed to build[/]'}"
            )
            console.print(
                f"[bold magenta]TrueFoundry:[/] [italic]{self.justification}[/]"
            )
            if not self.successful:
                console.print(f"[cyan]logs:[/] {self.logs}")

    def __init__(
        self,
        docker_client: docker.DockerClient,
        openai_client: OpenAI,
        environment: Dict,
    ):
        self.tools = [
            SendRequest(),
            Ask(),
            DockerRun(docker_client=docker_client, environment=environment),
        ]
        self.openai_client = openai_client



from truefoundry.ml._autogen.client.models import (  # type: ignore[attr-defined]
    InferMethodName,
    ModelVersionEnvironment,
    SklearnModelSchema,
    XGBoostModelSchema,
)
from truefoundry.ml._autogen.entities.artifacts import LibraryName
from truefoundry.ml.enums import (
    DataSlice,
    FileFormat,
    ModelFramework,
    ModelType,
    ViewType,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types import Image, Plot
from truefoundry.ml.log_types.artifacts.artifact import (
    ArtifactPath,
    ArtifactVersion,
    BlobStorageDirectory,
)
from truefoundry.ml.log_types.artifacts.dataset import DataDirectory, DataDirectoryPath
from truefoundry.ml.log_types.artifacts.model import ModelVersion
from truefoundry.ml.logger import init_logger
from truefoundry.ml.mlfoundry_api import get_client
from truefoundry.ml.mlfoundry_run import MlFoundryRun
from truefoundry.ml.model_framework import (
    FastAIFramework,
    GluonFramework,
    H2OFramework,
    KerasFramework,
    LightGBMFramework,
    ModelFrameworkType,
    ONNXFramework,
    PaddleFramework,
    PyTorchFramework,
    SklearnFramework,
    SpaCyFramework,
    StatsModelsFramework,
    TensorFlowFramework,
    TransformersFramework,
    XGBoostFramework,
    sklearn_infer_schema,
    xgboost_infer_schema,
)

__all__ = [
    "ArtifactPath",
    "ArtifactVersion",
    "BlobStorageDirectory",
    "DataDirectory",
    "DataDirectoryPath",
    "DataSlice",
    "FastAIFramework",
    "FileFormat",
    "get_client",
    "GluonFramework",
    "H2OFramework",
    "Image",
    "InferMethodName",
    "KerasFramework",
    "LibraryName",
    "LightGBMFramework",
    "MlFoundryException",
    "MlFoundryRun",
    "ModelFramework",
    "ModelFrameworkType",
    "ModelType",
    "ModelVersion",
    "ModelVersionEnvironment",
    "ONNXFramework",
    "PaddleFramework",
    "Plot",
    "PyTorchFramework",
    "sklearn_infer_schema",
    "SklearnFramework",
    "SklearnModelSchema",
    "SpaCyFramework",
    "StatsModelsFramework",
    "TensorFlowFramework",
    "TransformersFramework",
    "ViewType",
    "xgboost_infer_schema",
    "XGBoostFramework",
    "XGBoostModelSchema",
]

init_logger()



import functools
import os
import platform
import re
import time
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import urljoin, urlsplit

from truefoundry import client, version
from truefoundry.ml import constants
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ArtifactType,
    DeleteRunRequest,
    ListArtifactVersionsRequestDto,
    ListModelVersionsRequestDto,
    LogBatchRequestDto,
    MetricDto,
    MetricsApi,
    MlfoundryArtifactsApi,
    ModelVersionEnvironment,
    ParamDto,
    RunDataDto,
    RunDto,
    RunInfoDto,
    RunsApi,
    RunTagDto,
    UpdateRunRequestDto,
)
from truefoundry.ml.entities import Metric
from truefoundry.ml.enums import ModelFramework, RunStatus
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.internal_namespace import NAMESPACE
from truefoundry.ml.log_types import Image, Plot
from truefoundry.ml.log_types.artifacts.artifact import (
    ArtifactPath,
    ArtifactVersion,
    BlobStorageDirectory,
)
from truefoundry.ml.log_types.artifacts.general_artifact import _log_artifact_version
from truefoundry.ml.log_types.artifacts.model import (
    ModelVersion,
    _log_model_version,
)
from truefoundry.ml.logger import logger
from truefoundry.ml.run_utils import ParamsType, flatten_dict, process_params
from truefoundry.ml.session import ACTIVE_RUNS, _get_api_client
from truefoundry.ml.validation_utils import (
    MAX_ENTITY_KEY_LENGTH,
    MAX_METRICS_PER_BATCH,
    MAX_PARAMS_TAGS_PER_BATCH,
    _validate_batch_log_data,
)

if TYPE_CHECKING:
    import matplotlib
    import plotly

    from truefoundry.ml import ModelFrameworkType


def _ensure_not_deleted(method):
    @functools.wraps(method)
    def _check_deleted_or_not(self: "MlFoundryRun", *args, **kwargs):
        if self._deleted:
            raise MlFoundryException("Run was deleted, cannot access a deleted Run")
        else:
            return method(self, *args, **kwargs)

    return _check_deleted_or_not


class MlFoundryRun:
    """MlFoundryRun."""

    VALID_PARAM_AND_METRIC_NAMES = re.compile(r"^[A-Za-z0-9_\-\. /]+$")

    def __init__(
        self,
        experiment_id: str,
        run_id: str,
        auto_end: bool = False,
        **kwargs,
    ):
        """__init__.

        Args:
            experiment_id (str): experiment_id
            run_id (str): run_id
            auto_end (bool): If to end the run at garbage collection or process end (atexit)
        """
        self._experiment_id = str(experiment_id)
        self._run_id = run_id
        self._auto_end = auto_end
        self._run_info: Optional[RunInfoDto] = None
        self._run_data: Optional[RunDataDto] = None
        self._deleted = False
        self._terminate_called = False
        if self._auto_end:
            ACTIVE_RUNS.add_run(self)

        self._api_client = _get_api_client()
        self._runs_api = RunsApi(api_client=self._api_client)
        self._metrics_api = MetricsApi(api_client=self._api_client)
        self._mlfoundry_artifacts_api = MlfoundryArtifactsApi(
            api_client=self._api_client
        )

    @classmethod
    def _from_dto(cls, run_dto: RunDto) -> "MlFoundryRun":
        """classmethod to get MLfoundry run from dto instance"""
        run = cls(run_dto.info.experiment_id, run_dto.info.run_id)
        run._run_info = run_dto.info
        run._run_data = run_dto.data
        return run

    def _get_run_info(self) -> RunInfoDto:
        if self._run_info is not None:
            return self._run_info

        _run = self._runs_api.get_run_get(run_id=self.run_id)
        self._run_info = _run.run.info
        return self._run_info

    @property
    @_ensure_not_deleted
    def run_id(self) -> str:
        """Get run_id for the current `run`"""
        return self._run_id

    @property
    @_ensure_not_deleted
    def run_name(self) -> str:
        """Get run_name for the current `run`"""
        run_info = self._get_run_info()
        return run_info.name

    @property
    @_ensure_not_deleted
    def fqn(self) -> str:
        """Get fqn for the current `run`"""
        run_info = self._get_run_info()
        return run_info.fqn

    @property
    @_ensure_not_deleted
    def status(self) -> RunStatus:
        """Get status for the current `run`"""
        _run = self._runs_api.get_run_get(run_id=self.run_id)
        assert _run.run.info.status is not None
        return RunStatus(_run.run.info.status)

    @property
    @_ensure_not_deleted
    def ml_repo(self) -> str:
        """Get ml_repo name of which the current `run` is part of"""
        _experiment = client.ml_repos.get(id=self._experiment_id)
        return _experiment.data.manifest.name

    @property
    @_ensure_not_deleted
    def auto_end(self) -> bool:
        """Tells whether automatic end for `run` is True or False"""
        return self._auto_end

    @_ensure_not_deleted
    def __repr__(self) -> str:
        return f"<{type(self).__name__} at 0x{id(self):x}: run={self.fqn!r}>"

    @_ensure_not_deleted
    def __enter__(self):
        return self

    def _terminate_run_if_running(self, termination_status: RunStatus):
        """_terminate_run_if_running.

        Args:
            termination_status (RunStatus): termination_status
        """
        if self._terminate_called:
            return

        # Prevent double execution for termination
        self._terminate_called = True
        ACTIVE_RUNS.remove_run(self)

        current_status = self.status
        try:
            # we do not need to set any termination status unless the run was in RUNNING state
            if current_status != RunStatus.RUNNING:
                return
            logger.info("Setting run status of %r to %r", self.fqn, termination_status)
            _run_update = self._runs_api.update_run_post(
                update_run_request_dto=UpdateRunRequestDto(
                    run_id=self.run_id,
                    status=termination_status.value,
                    end_time=int(time.time() * 1000),
                )
            )
            self._run_info = _run_update.run_info
        except Exception as e:
            logger.warning(
                f"failed to set termination status {termination_status} due to {e}"
            )
        logger.info(f"Finished run: {self.fqn!r}, Dashboard: {self.dashboard_link}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
        self._terminate_run_if_running(status)

    def __del__(self):
        if self._auto_end:
            self._terminate_run_if_running(RunStatus.FINISHED)

    @property
    @_ensure_not_deleted
    def dashboard_link(self) -> str:
        """Get Mlfoundry dashboard link for a `run`"""
        tfy_host = "{uri.scheme}://{uri.netloc}/".format(
            uri=urlsplit(self._api_client.tfy_host)
        )
        return urljoin(tfy_host, f"mlfoundry/{self._experiment_id}/run/{self.run_id}/")

    @_ensure_not_deleted
    def end(self, status: RunStatus = RunStatus.FINISHED):
        """End a `run`.

        This function marks the run as `FINISHED`.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )
            # ...
            # Model training code
            # ...
            run.end()
            ```

            In case the run was created using the context manager approach,
            We do not need to call this function.

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            with client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            ) as run:
                # ...
                # Model training code
                ...
            # `run` will be automatically marked as `FINISHED` or `FAILED`.
            ```
        """
        self._terminate_run_if_running(status)

    @_ensure_not_deleted
    def delete(self) -> None:
        """
        This function permanently delete the run

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            client.create_ml_repo('iris-learning')
            run = client.create_run(ml_repo="iris-learning", run_name="svm-model1")
            run.log_params({"learning_rate": 0.001})
            run.log_metrics({"accuracy": 0.7, "loss": 0.6})

            run.delete()
            ```

            In case we try to call or access any other function of that run after deleting
            then it will through MlfoundryException

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            client.create_ml_repo('iris-learning')
            run = client.create_run(ml_repo="iris-learning", run_name="svm-model1")
            run.log_params({"learning_rate": 0.001})
            run.log_metrics({"accuracy": 0.7, "loss": 0.6})

            run.delete()
            run.log_params({"learning_rate": 0.001})  # raises MlfoundryException
            ```
        """
        name = self.run_name
        try:
            self._runs_api.hard_delete_run_post(
                delete_run_request=DeleteRunRequest(run_id=self.run_id)
            )
            logger.info(f"Run {name} was deleted successfully")
            ACTIVE_RUNS.remove_run(self)
            self._deleted = True
            self._auto_end = False
        except Exception as ex:
            logger.warning(f"Failed to delete the run {name} because of {ex}")
            raise

    @_ensure_not_deleted
    def list_artifact_versions(
        self,
        artifact_type: Optional[ArtifactType] = ArtifactType.ARTIFACT,
    ) -> Iterator[ArtifactVersion]:
        """
        Get all the version of an artifact from a particular run to download contents or load them in memory

        Args:
            artifact_type: Type of the artifact you want

        Returns:
            Iterator[ArtifactVersion]: An iterator that yields non deleted artifact-versions
                of an artifact under a given run  sorted reverse by the version number

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(ml_repo="iris-learning", run_name="svm-model1")
            artifact_versions = run.list_artifact_versions()

            for artifact_version in artifact_versions:
                print(artifact_version)

            run.end()
            ```
        """
        done, page_token, max_results = False, None, 25
        while not done:
            _artifact_versions = (
                self._mlfoundry_artifacts_api.list_artifact_versions_post(
                    list_artifact_versions_request_dto=ListArtifactVersionsRequestDto(
                        run_ids=[self.run_id],
                        artifact_types=[artifact_type] if artifact_type else None,
                        max_results=max_results,
                        page_token=page_token,
                    )
                )
            )
            artifact_versions = _artifact_versions.artifact_versions
            page_token = _artifact_versions.next_page_token
            for artifact_version in artifact_versions:
                yield ArtifactVersion.from_fqn(artifact_version.fqn)
            if not artifact_versions or page_token is None:
                done = True

    @_ensure_not_deleted
    def list_model_versions(
        self,
    ) -> Iterator[ModelVersion]:
        """
        Get all the version of a models from a particular run to download contents or load them in memory

        Returns:
            Iterator[ModelVersion]: An iterator that yields non deleted model-versions
                under a given run  sorted reverse by the version number

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.get_run(run_id="<your-run-id>")
            model_versions = run.list_model_versions()

            for model_version in model_versions:
                print(model_version)

            run.end()
            ```
        """
        done, page_token, max_results = False, None, 25
        while not done:
            _model_versions = self._mlfoundry_artifacts_api.list_model_versions_post(
                list_model_versions_request_dto=ListModelVersionsRequestDto(
                    run_ids=[self.run_id],
                    max_results=max_results,
                    page_token=page_token,
                )
            )
            model_versions = _model_versions.model_versions
            page_token = _model_versions.next_page_token
            for model_version in model_versions:
                yield ModelVersion.from_fqn(fqn=model_version.fqn)
            if not model_versions or page_token is None:
                done = True

    def _add_git_info(self, root_path: Optional[str] = None):
        """_add_git_info.

        Args:
            root_path (Optional[str]): root_path
        """
        root_path = root_path or os.getcwd()
        try:
            from truefoundry.ml.git_info import GitInfo

            git_info = GitInfo(root_path)
            tags = [
                RunTagDto(
                    key=constants.GIT_COMMIT_TAG_NAME,
                    value=git_info.current_commit_sha,
                ),
                RunTagDto(
                    key=constants.GIT_BRANCH_TAG_NAME,
                    value=git_info.current_branch_name,
                ),
                RunTagDto(
                    key=constants.GIT_DIRTY_TAG_NAME, value=str(git_info.is_dirty)
                ),
            ]
            remote_url = git_info.remote_url
            if remote_url is not None:
                tags.append(
                    RunTagDto(key=constants.GIT_REMOTE_URL_NAME, value=remote_url)
                )
            _validate_batch_log_data(metrics=[], params=[], tags=tags)
            self._runs_api.log_run_batch_post(
                log_batch_request_dto=LogBatchRequestDto(run_id=self.run_id, tags=tags)
            )
        except Exception as ex:
            # no-blocking
            logger.warning(f"failed to log git info because {ex}")

    def _add_python_truefoundry_version(self):
        python_version = platform.python_version()
        truefoundry_version = version.__version__
        tags = [
            RunTagDto(
                key=constants.PYTHON_VERSION_TAG_NAME,
                value=python_version,
            ),
        ]

        if truefoundry_version:
            tags.append(
                RunTagDto(
                    key=constants.MLFOUNDRY_VERSION_TAG_NAME,
                    value=truefoundry_version,
                )
            )
            tags.append(
                RunTagDto(
                    key=constants.TRUEFOUNDRY_VERSION_TAG_NAME,
                    value=truefoundry_version,
                )
            )
        else:
            logger.warning("Failed to get MLFoundry version.")
        _validate_batch_log_data(metrics=[], params=[], tags=tags)
        self._runs_api.log_run_batch_post(
            log_batch_request_dto=LogBatchRequestDto(run_id=self.run_id, tags=tags)
        )

    @_ensure_not_deleted
    def log_artifact(
        self,
        name: str,
        artifact_paths: List[
            Union[Tuple[str], Tuple[str, Optional[str]], ArtifactPath]
        ],
        description: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        step: Optional[int] = 0,
        progress: Optional[bool] = None,
    ) -> ArtifactVersion:
        """Logs an artifact for the current ML Repo.

        An `artifact` is a list of local files and directories.
        This function packs the mentioned files and directories in `artifact_paths`
        and uploads them to remote storage linked to the experiment

        Args:
            name (str): Name of the Artifact. If an artifact with this name already exists under the current ML Repo,
                the logged artifact will be added as a new version under that `name`.
                If no artifact exist with the given `name`, the given artifact will be logged as version 1.
            artifact_paths (List[truefoundry.ml.ArtifactPath], optional): A list of pairs
                of (source path, destination path) to add files and folders
                to the artifact version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the artifact contents to upload to.

                ```python
                from truefoundry.ml import ArtifactPath

                ...
                run.log_artifact(
                     name="xyz",
                     artifact_paths=[
                        ArtifactPath("foo.txt", "foo/bar/foo.txt"),
                        ArtifactPath("tokenizer/", "foo/tokenizer/"),
                        ArtifactPath('bar.text'),
                        ('bar.txt', ),
                        ('foo.txt', 'a/foo.txt')
                     ]
                )
                ```

                would result in

                ```
                .
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            step (int): step/iteration at which the vesion is being logged, defaults to 0.
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            truefoundry.ml.ArtifactVersion: an instance of `ArtifactVersion` that can be used to download the files,
            or update attributes like description, metadata.

        Examples:

            ```python
            import os
            from truefoundry.ml import get_client, ArtifactPath

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )

            run.log_artifact(
                name="hello-world-file",
                artifact_paths=[ArtifactPath('artifact.txt', 'a/b/')]
            )

            run.end()
            ```
        """
        if not artifact_paths:
            raise MlFoundryException(
                "artifact_paths cannot be empty, atleast one artifact_path must be passed"
            )

        return _log_artifact_version(
            run=self,
            name=name,
            artifact_paths=artifact_paths,
            description=description,
            metadata=metadata,
            step=step,
            progress=progress,
        )

    @_ensure_not_deleted
    def log_metrics(self, metric_dict: Dict[str, Union[int, float]], step: int = 0):
        """Log metrics for the current `run`.

        A metric is defined by a metric name (such as "training-loss") and a
        floating point or integral value (such as `1.2`). A metric is associated
        with a `step` which is the training iteration at which the metric was
        calculated.

        Args:
            metric_dict (Dict[str, Union[int, float]]): A metric name to metric value map.
                metric value should be either `float` or `int`. This should be
                a non-empty dictionary.
            step (int, optional): Training step/iteration at which the metrics
                present in `metric_dict` were calculated. If not passed, `0` is
                set as the `step`.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_metrics(metric_dict={"accuracy": 0.7, "loss": 0.6}, step=0)
            run.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4}, step=1)

            run.end()
            ```
        """
        timestamp = int(time.time() * 1000)
        metrics = []
        for key in metric_dict.keys():
            if isinstance(metric_dict[key], str):
                logger.warning(
                    f"Cannot log metric with string value. Discarding metric {key}={metric_dict[key]}"
                )
                continue
            if not self.VALID_PARAM_AND_METRIC_NAMES.match(key):
                logger.warning(
                    f"Invalid metric name: {key}. Names may only contain alphanumerics, "
                    f"underscores (_), dashes (-), periods (.), spaces ( ), and slashes (/). "
                    f"Discarding metric {key}={metric_dict[key]}"
                )
                continue
            metrics.append(
                MetricDto(
                    key=key, value=metric_dict[key], timestamp=timestamp, step=step
                )
            )

        if len(metrics) == 0:
            logger.warning("Cannot log empty metrics dictionary")
            return

        try:
            for i in range(0, len(metrics), MAX_METRICS_PER_BATCH):
                metrics_batch = metrics[i : i + MAX_METRICS_PER_BATCH]

                _validate_batch_log_data(metrics=metrics_batch, params=[], tags=[])

                self._runs_api.log_run_batch_post(
                    log_batch_request_dto=LogBatchRequestDto(
                        run_id=self.run_id, metrics=metrics_batch, params=[], tags=[]
                    )
                )
        except Exception as e:
            raise MlFoundryException(str(e)) from e

        logger.info("Metrics logged successfully")

    @_ensure_not_deleted
    def log_params(self, param_dict: ParamsType, flatten_params: bool = False):
        """Logs parameters for the run.

        Parameters or Hyperparameters can be thought of as configurations for a run.
        For example, the type of kernel used in a SVM model is a parameter.
        A Parameter is defined by a name and a string value. Parameters are
        also immutable, we cannot overwrite parameter value for a parameter
        name.

        Args:
            param_dict (ParamsType): A parameter name to parameter value map.
                Parameter values are converted to `str`.
            flatten_params (bool): Flatten hierarchical dict, e.g. `{'a': {'b': 'c'}} -> {'a.b': 'c'}`.
                All the keys will be converted to `str`. Defaults to False

        Examples:

            ### Logging parameters using a `dict`.

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_params({"learning_rate": 0.01, "epochs": 10})

            run.end()
            ```

            ### Logging parameters using `argparse` Namespace object

            ```python
            import argparse
            from truefoundry.ml import get_client

            parser = argparse.ArgumentParser()
            parser.add_argument("-batch_size", type=int, required=True)
            args = parser.parse_args()

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_params(args)
            ```
        """
        try:
            param_dict = process_params(param_dict)
            param_dict = flatten_dict(param_dict) if flatten_params else param_dict

            params = []
            for param_key in param_dict.keys():
                if (
                    len(str(param_key)) > MAX_ENTITY_KEY_LENGTH
                    or len(str(param_dict[param_key])) > MAX_ENTITY_KEY_LENGTH
                ):
                    logger.warning(
                        f"MlFoundry can't log parameters with length greater than {MAX_ENTITY_KEY_LENGTH} characters. "
                        f"Discarding {param_key}:{param_dict[param_key]}."
                    )
                    continue
                if not self.VALID_PARAM_AND_METRIC_NAMES.match(param_key):
                    logger.warning(
                        f"Invalid param name: {param_key}. Names may only contain alphanumerics, "
                        f"underscores (_), dashes (-), periods (.), spaces ( ), and slashes (/). "
                        f"Discarding param {param_key}={param_dict[param_key]}"
                    )
                    continue
                params.append(ParamDto(key=param_key, value=str(param_dict[param_key])))

            if len(params) == 0:
                logger.warning("Cannot log empty params dictionary")

            for i in range(0, len(params), MAX_PARAMS_TAGS_PER_BATCH):
                params_batch = params[i : i + MAX_PARAMS_TAGS_PER_BATCH]

                _validate_batch_log_data(metrics=[], params=params_batch, tags=[])
                logger.debug("Logging parameters: %s", params_batch)

                self._runs_api.log_run_batch_post(
                    log_batch_request_dto=LogBatchRequestDto(
                        run_id=self.run_id, metrics=[], params=params_batch, tags=[]
                    )
                )
        except Exception as e:
            raise MlFoundryException(str(e)) from e
        logger.info("Parameters logged successfully")

    @_ensure_not_deleted
    def set_tags(self, tags: Dict[str, str]):
        """Set tags for the current `run`.

        Tags are "labels" for a run. A tag is represented by a tag name and value.

        Args:
            tags (Dict[str, str]): A tag name to value map.
                Tag name cannot start with `mlf.`, `mlf.` prefix
                is reserved for truefoundry. Tag values will be converted
                to `str`.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.set_tags({"nlp.framework": "Spark NLP"})

            run.end()
            ```
        """
        tags = tags or {}
        try:
            NAMESPACE.validate_namespace_not_used(names=tags.keys())
            tags_arr = [
                RunTagDto(key=key, value=str(value)) for key, value in tags.items()
            ]
            for i in range(0, len(tags_arr), MAX_PARAMS_TAGS_PER_BATCH):
                tags_batch = tags_arr[i : i + MAX_PARAMS_TAGS_PER_BATCH]

                _validate_batch_log_data(metrics=[], params=[], tags=tags_batch)
                self._runs_api.log_run_batch_post(
                    log_batch_request_dto=LogBatchRequestDto(
                        run_id=self.run_id, metrics=[], params=[], tags=tags_batch
                    )
                )
        except Exception as e:
            raise MlFoundryException(str(e)) from e
        logger.info("Tags set successfully")

    @_ensure_not_deleted
    def get_tags(self, no_cache=False) -> Dict[str, str]:
        """Returns all the tags set for the current `run`.

        Returns:
            Dict[str, str]: A dictionary containing tags. The keys in the dictionary
                are tag names and the values are corresponding tag values.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.set_tags({"nlp.framework": "Spark NLP"})
            print(run.get_tags())

            run.end()
            ```
        """
        if no_cache or not self._run_data:
            _run = self._runs_api.get_run_get(run_id=self.run_id)
            self._run_data = _run.run.data
        assert self._run_data is not None
        tags = self._run_data.tags or []
        return {tag.key: tag.value for tag in tags}

    @_ensure_not_deleted
    def get_metrics(
        self, metric_names: Optional[Iterable[str]] = None
    ) -> Dict[str, List[Metric]]:
        """Get metrics logged for the current `run` grouped by metric name.

        Args:
            metric_names (Optional[Iterable[str]], optional): A list of metric names
                For which the logged metrics will be fetched. If not passed, then all
                metrics logged under the `run` is returned.

        Returns:
            Dict[str, List[Metric]]: A dictionary containing metric name to list of metrics
                map.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            )
            run.log_metrics(metric_dict={"accuracy": 0.7, "loss": 0.6}, step=0)
            run.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4}, step=1)

            metrics = run.get_metrics()
            for metric_name, metric_history in metrics.items():
                print(f"logged metrics for metric {metric_name}:")
                for metric in metric_history:
                    print(f"value: {metric.value}")
                    print(f"step: {metric.step}")
                    print(f"timestamp_ms: {metric.timestamp}")
                    print("--")

            run.end()
            ```
        """
        _run = self._runs_api.get_run_get(run_id=self.run_id)
        assert _run.run.data is not None
        run_metrics = _run.run.data.metrics or []
        run_metric_names = {metric.key for metric in run_metrics}

        metric_names = (
            set(metric_names) if metric_names is not None else run_metric_names
        )

        unknown_metrics = metric_names - run_metric_names
        if len(unknown_metrics) > 0:
            logger.warning(f"{unknown_metrics} metrics not present in the run")
        metrics_dict: Dict[str, List[Metric]] = {
            metric_name: [] for metric_name in unknown_metrics
        }
        valid_metrics = metric_names - unknown_metrics
        for metric_name in valid_metrics:
            _metric_history = self._metrics_api.get_metric_history_get(
                run_id=self.run_id, metric_key=metric_name
            )
            metrics_dict[metric_name] = [
                Metric.from_dto(metric) for metric in _metric_history.metrics
            ]
        return metrics_dict

    @_ensure_not_deleted
    def get_params(self, no_cache=False) -> Dict[str, str]:
        """Get all the params logged for the current `run`.

        Returns:
            Dict[str, str]: A dictionary containing the parameters. The keys in the dictionary
                are parameter names and the values are corresponding parameter values.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            run.log_params({"learning_rate": 0.01, "epochs": 10})
            print(run.get_params())

            run.end()
            ```
        """
        if no_cache or not self._run_data:
            _run = self._runs_api.get_run_get(run_id=self.run_id)
            self._run_data = _run.run.data
        assert self._run_data is not None
        params = self._run_data.params or []
        return {param.key: param.value for param in params}

    @_ensure_not_deleted
    def log_model(
        self,
        *,
        name: str,
        model_file_or_folder: Union[str, BlobStorageDirectory],
        description: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        step: int = 0,
        progress: Optional[bool] = None,
        framework: Optional[Union[str, ModelFramework, "ModelFrameworkType"]] = None,
        environment: Optional[ModelVersionEnvironment] = None,
    ) -> ModelVersion:
        # TODO (chiragjn): Document mapping of framework to list of valid model save kwargs
        # TODO (chiragjn): Add more examples
        """
        Serialize and log a versioned model under the current ML Repo. Each logged model generates a new version
            associated with the given `name` and linked to the current run. Multiple versions of the model can be
            logged as separate versions under the same `name`.

        Args:
            name (str): Name of the model. If a model with this name already exists under the current ML Repo,
                the logged model will be added as a new version under that `name`. If no models exist with the given
                `name`, the given model will be logged as version 1.
            model_file_or_folder (Union[str, BlobStorageDirectory]):
                str:
                    Path to either a single file or a folder containing model files.
                    This folder is typically created using serialization methods from libraries or frameworks,
                    e.g., `joblib.dump`, `model.save_pretrained(...)`, `torch.save(...)`, or `model.save(...)`.
                BlobStorageDirectory:
                    uri (str): URI to the model file or folder in a storage integration associated with the specified ML Repo.
                        The model files or folder must reside within the same storage integration as the specified ML Repo.
                        Accepted URI formats include `s3://integration-bucket-name/prefix/path/to/model` or `gs://integration-bucket-name/prefix/path/to/model`.
                        If the URI points to a model in a different storage integration, an error will be raised indicating "Invalid source URI."
            framework (Optional[Union[ModelFramework, ModelFrameworkType]]): Framework used for model serialization.
                Supported frameworks values (ModelFrameworkType) can be imported from `from truefoundry.ml import *`.
                Supported frameworks can be found in `truefoundry.ml.enums.ModelFramework`.
                Can also be `None` if the framework is not known or not supported.
                **Deprecated**: Prefer `ModelFrameworkType` over `enums.ModelFramework`.

            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            step (int): step/iteration at which the model is being logged, defaults to 0.
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            truefoundry.ml.ModelVersion: an instance of `ModelVersion` that can be used to download the files,
                load the model, or update attributes like description, metadata, schema.

        Examples:

            ### Sklearn

            ```python
            from truefoundry.ml import get_client
            from truefoundry.ml.enums import ModelFramework

            import joblib
            import numpy as np
            from sklearn.pipeline import make_pipeline
            from sklearn.preprocessing import StandardScaler
            from sklearn.svm import SVC

            X = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
            y = np.array([1, 1, 2, 2])
            clf = make_pipeline(StandardScaler(), SVC(gamma='auto'))
            clf.fit(X, y)
            joblib.dump(clf, "sklearn-pipeline.joblib")

            client = get_client()
            client.create_ml_repo(  # This is only required once
                ml_repo="my-classification-project",
                # This controls which bucket is used.
                # You can get this from Integrations > Blob Storage. `None` picks the default
                storage_integration_fqn=None
            )
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            model_version = run.log_model(
                name="my-sklearn-model",
                model_file_or_folder="sklearn-pipeline.joblib",
                framework=ModelFramework.SKLEARN,
                metadata={"accuracy": 0.99, "f1": 0.80},
                step=1,  # step number, useful when using iterative algorithms like SGD
            )
            print(model_version.fqn)
            ```

            ### Huggingface Transformers

            ```python
            from truefoundry.ml import get_client
            from truefoundry.ml.enums import ModelFramework

            import torch
            from transformers import AutoTokenizer, AutoConfig, pipeline, AutoModelForCausalLM
            pln = pipeline(
                "text-generation",
                model_file_or_folder="EleutherAI/pythia-70m",
                tokenizer="EleutherAI/pythia-70m",
                torch_dtype=torch.float16
            )
            pln.model.save_pretrained("my-transformers-model")
            pln.tokenizer.save_pretrained("my-transformers-model")

            client = get_client()
            client.create_ml_repo(  # This is only required once
                ml_repo="my-llm-project",
                # This controls which bucket is used.
                # You can get this from Integrations > Blob Storage. `None` picks the default
                storage_integration_fqn=None
            )
            run = client.create_run(
                ml_repo="my-llm-project"
            )
            model_version = run.log_model(
                name="my-transformers-model",
                model_file_or_folder="my-transformers-model/",
                framework=ModelFramework.TRANSFORMERS
            )
            print(model_version.fqn)
            ```
        """

        model_version = _log_model_version(
            run=self,
            name=name,
            model_file_or_folder=model_file_or_folder,
            description=description,
            metadata=metadata,
            step=step,
            progress=progress,
            framework=framework,
            environment=environment,
        )
        logger.info(f"Logged model successfully with fqn {model_version.fqn!r}")
        return model_version

    @_ensure_not_deleted
    def log_images(self, images: Dict[str, Image], step: int = 0):
        """Log images under the current `run` at the given `step`.

        Use this function to log images for a `run`. `PIL` package is needed to log images.
        To install the `PIL` package, run `pip install pillow`.

        Args:
            images (Dict[str, "truefoundry.ml.Image"]): A map of string image key to instance of
                `truefoundry.ml.Image` class. The image key should only contain alphanumeric,
                hyphens(-) or underscores(_). For a single key and step pair, we can log only
                one image.
            step (int, optional): Training step/iteration for which the `images` should be
                logged. Default is `0`.

        Examples:

            ### Logging images from different sources

            ```python
            from truefoundry.ml import get_client, Image
            import numpy as np
            import PIL.Image

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project",
            )

            imarray = np.random.randint(low=0, high=256, size=(100, 100, 3))
            im = PIL.Image.fromarray(imarray.astype("uint8")).convert("RGB")
            im.save("result_image.jpeg")

            images_to_log = {
                "logged-image-array": Image(data_or_path=imarray),
                "logged-pil-image": Image(data_or_path=im),
                "logged-image-from-path": Image(data_or_path="result_image.jpeg"),
            }

            run.log_images(images_to_log, step=1)
            run.end()
            ```
        """
        for key, image in images.items():
            if not isinstance(image, Image):
                raise MlFoundryException(
                    "image should be of type `truefoundry.ml.Image`"
                )
            image.save(run=self, key=key, step=step)

    @_ensure_not_deleted
    def log_plots(
        self,
        plots: Dict[  # type: ignore[valid-type]
            str,
            Union[
                "matplotlib.pyplot",
                "matplotlib.figure.Figure",
                "plotly.graph_objects.Figure",
                Plot,
            ],
        ],
        step: int = 0,
    ):
        """Log custom plots under the current `run` at the given `step`.

        Use this function to log custom matplotlib, plotly plots.

        Args:
            plots (Dict[str, "matplotlib.pyplot", "matplotlib.figure.Figure", "plotly.graph_objects.Figure", Plot]):
                A map of string plot key to the plot or figure object.
                The plot key should only contain alphanumeric, hyphens(-) or
                underscores(_). For a single key and step pair, we can log only
                one image.
            step (int, optional): Training step/iteration for which the `plots` should be
                logged. Default is `0`.


        Examples:

            ### Logging a plotly figure

            ```python
            from truefoundry.ml import get_client
            import plotly.express as px

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project",
            )

            df = px.data.tips()
            fig = px.histogram(
                df,
                x="total_bill",
                y="tip",
                color="sex",
                marginal="rug",
                hover_data=df.columns,
            )

            plots_to_log = {
                "distribution-plot": fig,
            }

            run.log_plots(plots_to_log, step=1)
            run.end()
            ```


            ### Logging a matplotlib plt or figure

            ```python
            from truefoundry.ml import get_client
            from matplotlib import pyplot as plt
            import numpy as np

            client = get_client()
            run = client.create_run(
                ml_repo="my-classification-project",
            )

            t = np.arange(0.0, 5.0, 0.01)
            s = np.cos(2 * np.pi * t)
            (line,) = plt.plot(t, s, lw=2)

            plt.annotate(
                "local max",
                xy=(2, 1),
                xytext=(3, 1.5),
                arrowprops=dict(facecolor="black", shrink=0.05),
            )

            plt.ylim(-2, 2)

            plots_to_log = {"cos-plot": plt, "cos-plot-using-figure": plt.gcf()}

            run.log_plots(plots_to_log, step=1)
            run.end()
            ```
        """
        for key, plot in plots.items():
            plot = Plot(plot) if not isinstance(plot, Plot) else plot
            plot.save(run=self, key=key, step=step)



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.api_client import ApiClient
from truefoundry.ml._autogen.client.api_response import ApiResponse
from truefoundry.ml._autogen.client.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError,
)
from truefoundry.ml._autogen.client.models.get_metric_history_response import (
    GetMetricHistoryResponse,
)
from truefoundry.ml._autogen.client.models.list_metric_history_request_dto import (
    ListMetricHistoryRequestDto,
)
from truefoundry.ml._autogen.client.models.list_metric_history_response_dto import (
    ListMetricHistoryResponseDto,
)
from truefoundry.pydantic_v1 import StrictStr, validate_arguments


class MetricsApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def get_metric_history_get(
        self,
        run_uuid: Optional[StrictStr] = None,
        run_id: Optional[StrictStr] = None,
        metric_key: Optional[StrictStr] = None,
        **kwargs,
    ) -> GetMetricHistoryResponse:  # noqa: E501
        """Get Metric History  # noqa: E501

        Get Metric History  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_metric_history_get(run_uuid, run_id, metric_key, async_req=True)
        >>> result = thread.get()

        :param run_uuid:
        :type run_uuid: str
        :param run_id:
        :type run_id: str
        :param metric_key:
        :type metric_key: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: GetMetricHistoryResponse
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the get_metric_history_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.get_metric_history_get_with_http_info(
            run_uuid, run_id, metric_key, **kwargs
        )  # noqa: E501

    @validate_arguments
    def get_metric_history_get_with_http_info(
        self,
        run_uuid: Optional[StrictStr] = None,
        run_id: Optional[StrictStr] = None,
        metric_key: Optional[StrictStr] = None,
        **kwargs,
    ) -> ApiResponse:  # noqa: E501
        """Get Metric History  # noqa: E501

        Get Metric History  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_metric_history_get_with_http_info(run_uuid, run_id, metric_key, async_req=True)
        >>> result = thread.get()

        :param run_uuid:
        :type run_uuid: str
        :param run_id:
        :type run_id: str
        :param metric_key:
        :type metric_key: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(GetMetricHistoryResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["run_uuid", "run_id", "metric_key"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_metric_history_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("run_uuid") is not None:  # noqa: E501
            _query_params.append(("run_uuid", _params["run_uuid"]))

        if _params.get("run_id") is not None:  # noqa: E501
            _query_params.append(("run_id", _params["run_id"]))

        if _params.get("metric_key") is not None:  # noqa: E501
            _query_params.append(("metric_key", _params["metric_key"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "GetMetricHistoryResponse",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/metrics/get-history",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def list_metric_history_post(
        self, list_metric_history_request_dto: ListMetricHistoryRequestDto, **kwargs
    ) -> ListMetricHistoryResponseDto:  # noqa: E501
        """List Metric History  # noqa: E501

        List Metric History  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_metric_history_post(list_metric_history_request_dto, async_req=True)
        >>> result = thread.get()

        :param list_metric_history_request_dto: (required)
        :type list_metric_history_request_dto: ListMetricHistoryRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ListMetricHistoryResponseDto
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the list_metric_history_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.list_metric_history_post_with_http_info(
            list_metric_history_request_dto, **kwargs
        )  # noqa: E501

    @validate_arguments
    def list_metric_history_post_with_http_info(
        self, list_metric_history_request_dto: ListMetricHistoryRequestDto, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """List Metric History  # noqa: E501

        List Metric History  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_metric_history_post_with_http_info(list_metric_history_request_dto, async_req=True)
        >>> result = thread.get()

        :param list_metric_history_request_dto: (required)
        :type list_metric_history_request_dto: ListMetricHistoryRequestDto
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ListMetricHistoryResponseDto, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["list_metric_history_request_dto"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_metric_history_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["list_metric_history_request_dto"] is not None:
            _body_params = _params["list_metric_history_request_dto"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["APIKey"]  # noqa: E501

        _response_types_map = {
            "200": "ListMetricHistoryResponseDto",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/api/2.0/mlflow/metrics/list-history",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )



from truefoundry.deploy._autogen.models import (
    AppProtocol,
    Autoshutdown,
    CanaryStep,
    CapacityType,
    Claim,
    ConcurrencyPolicy,
    JobAlert,
    Kustomize,
    ParamType,
    Protocol,
    SparkDriverConfig,
    SparkExecutorConfig,
    SparkExecutorDynamicScaling,
    SparkExecutorFixedInstances,
    WorkbenchImage,
    WorkflowAlert,
)
from truefoundry.deploy.lib.dao.application import (
    delete_application,
    get_application,
    get_job_run,
    list_applications,
    list_job_runs,
    terminate_job_run,
    trigger_job,
    trigger_workflow,
)
from truefoundry.deploy.lib.dao.version import (
    get_version as get_application_version,
)
from truefoundry.deploy.lib.dao.version import (
    list_versions as list_application_versions,
)
from truefoundry.deploy.lib.dao.workspace import (
    delete_workspace,
    get_workspace_by_fqn,
    list_workspaces,
)
from truefoundry.deploy.v2.lib.deployable_patched_models import (
    Application,
    ApplicationSet,
    AsyncService,
    Helm,
    Job,
    Notebook,
    RStudio,
    Service,
    SparkJob,
    SSHServer,
    Volume,
    Workflow,
)
from truefoundry.deploy.v2.lib.patched_models import (
    UV,
    AMQPInputConfig,
    AMQPMetricConfig,
    AMQPOutputConfig,
    ArtifactsCacheVolume,
    ArtifactsDownload,
    AsyncProcessorSidecar,
    AsyncServiceAutoscaling,
    Autoscaling,
    AWSAccessKeyAuth,
    AWSInferentia,
    AWSInferentiaAccelerator,
    AwsSqsAccessKeyBasedAuth,
    AwsSqsAssumedRoleBasedAuth,
    BasicAuthCreds,
    BlueGreen,
    Build,
    Canary,
    CoreNATSOutputConfig,
    CPUUtilizationMetric,
    CronMetric,
    CUDAVersion,
    DockerFileBuild,
    DynamicVolumeConfig,
    Email,
    Endpoint,
    GcpTPU,
    GitHelmRepo,
    GitSource,
    GPUType,
    HealthProbe,
    HelmRepo,
    HttpProbe,
    HuggingfaceArtifactSource,
    Image,
    JwtAuthConfig,
    KafkaInputConfig,
    KafkaMetricConfig,
    KafkaOutputConfig,
    KafkaSASLAuth,
    LocalSource,
    Manual,
    NATSInputConfig,
    NATSMetricConfig,
    NATSOutputConfig,
    NATSUserPasswordAuth,
    NodepoolSelector,
    NodeSelector,
    NvidiaGPU,
    NvidiaMIGGPU,
    NvidiaTimeslicingGPU,
    OCIRepo,
    Param,
    Pip,
    Poetry,
    Port,
    PythonBuild,
    RemoteSource,
    Resources,
    Rolling,
    RPSMetric,
    Schedule,
    SecretMount,
    ServiceAutoscaling,
    SlackBot,
    SlackWebhook,
    SparkBuild,
    SparkImage,
    SparkImageBuild,
    SparkJobJavaEntrypoint,
    SparkJobPythonEntrypoint,
    SparkJobPythonNotebookEntrypoint,
    SparkJobScalaEntrypoint,
    SparkJobScalaNotebookEntrypoint,
    SQSInputConfig,
    SQSOutputConfig,
    SQSQueueMetricConfig,
    StaticVolumeConfig,
    StringDataMount,
    TPUType,
    TrueFoundryArtifactSource,
    TruefoundryArtifactSource,  # deprecated, kept for backwards compatibility
    TrueFoundryInteractiveLogin,
    VolumeBrowser,
    VolumeMount,
    WorkerConfig,
)



import os
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml

DEFAULT_KUBECONFIG_PATH: Path = Path.home() / ".kube" / "config"
CLUSTER_NAME_FORMAT: str = "tfy-{cluster}-cluster"
USER_NAME_FORMAT: str = "tfy-{cluster}-user"
CONTEXT_NAME_FORMAT: str = "tfy-{cluster}-context"
KUBE_CONFIG_CONTENT = {
    "apiVersion": "v1",
    "kind": "Config",
    "clusters": [],
    "users": [],
    "contexts": [],
}


def get_kubeconfig_path() -> Path:
    """
    Returns the kubeconfig path to use.
    If KUBECONFIG is set, returns the first path from the environment variable.
    Otherwise, returns the default kubeconfig path.
    """
    kubeconfig_env = os.environ.get("KUBECONFIG")
    if kubeconfig_env:
        # Use the first path in KUBECONFIG if multiple are provided
        first_path = kubeconfig_env.split(os.pathsep)[0]
        return Path(first_path)
    return DEFAULT_KUBECONFIG_PATH


def get_kubeconfig_content(path: Path = DEFAULT_KUBECONFIG_PATH) -> Dict[str, Any]:
    if path.exists():
        with open(path, "r") as f:
            return yaml.safe_load(f) or KUBE_CONFIG_CONTENT
    else:
        return KUBE_CONFIG_CONTENT


def save_kubeconfig(config: Dict[str, Any], path: Path) -> None:
    config["apiVersion"] = "v1"
    config["kind"] = "Config"
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as f:
        yaml.safe_dump(config, f, default_flow_style=False)


def add_update_cluster_context(
    config: Dict[str, Any],
    cluster: str,
    server_url: str,
    exec_command: List[str],
    envs: Optional[Dict[str, str]] = None,
) -> str:
    """
    Adds a new cluster context to the given kubeconfig dictionary using exec-based authentication.
    """
    cluster_name: str = CLUSTER_NAME_FORMAT.format(cluster=cluster)
    user_name: str = USER_NAME_FORMAT.format(cluster=cluster)
    context_name: str = CONTEXT_NAME_FORMAT.format(cluster=cluster)

    # Add or update cluster
    config["clusters"] = [
        c for c in config.get("clusters", []) if c["name"] != cluster_name
    ]
    config["clusters"].append(
        {
            "name": cluster_name,
            "cluster": {"server": server_url},
        }
    )

    # Add or update user with exec command
    config["users"] = [u for u in config.get("users", []) if u["name"] != user_name]
    config["users"].append(
        {
            "name": user_name,
            "user": {
                "exec": {
                    "apiVersion": "client.authentication.k8s.io/v1beta1",
                    "command": exec_command[0],
                    "args": exec_command[1:],
                    "env": (
                        [{"name": k, "value": v} for k, v in envs.items()]
                        if envs
                        else []
                    ),
                }
            },
        }
    )

    # Add or update context
    config["contexts"] = [
        c for c in config.get("contexts", []) if c["name"] != context_name
    ]
    config["contexts"].append(
        {"name": context_name, "context": {"cluster": cluster_name, "user": user_name}}
    )
    return context_name


def get_cluster_context(
    config: Dict[str, Any], cluster: str
) -> Optional[Dict[str, Any]]:
    cluster_name: str = CLUSTER_NAME_FORMAT.format(cluster=cluster)
    return next(
        (c for c in config.get("clusters", []) if c["name"] == cluster_name), None
    )


def get_cluster_server_url(config: Dict[str, Any], cluster: str) -> Optional[str]:
    cluster_context: Optional[Dict[str, Any]] = get_cluster_context(config, cluster)
    if cluster_context:
        return cluster_context["cluster"].get("server")
    return None



try:
    import fsspec
    from flytekit import task as _
except ImportError:
    print(
        "To use workflows, please run 'pip install truefoundry[workflow]'. "
        "Note: The `workflow` feature is only available for Python 3.9 to 3.12"
    )

from flytekit import conditional
from flytekit.types.directory import FlyteDirectory
from flytekit.types.error.error import FlyteError
from flytekit.types.file import FlyteFile

from truefoundry.common.constants import ENV_VARS
from truefoundry.deploy.v2.lib.patched_models import (
    ContainerTaskConfig,
    DatabricksJobTaskConfig,
    PySparkTaskConfig,
    PythonTaskConfig,
    TaskDockerFileBuild,
    TaskPythonBuild,
)
from truefoundry.workflow.container_task import ContainerTask
from truefoundry.workflow.map_task import map_task
from truefoundry.workflow.python_task import PythonFunctionTask
from truefoundry.workflow.remote_filesystem.tfy_signed_url_fs import SignedURLFileSystem
from truefoundry.workflow.task import task
from truefoundry.workflow.workflow import ExecutionConfig, workflow

__all__ = [
    "task",
    "ContainerTask",
    "PythonFunctionTask",
    "map_task",
    "workflow",
    "conditional",
    "FlyteDirectory",
    "TaskDockerFileBuild",
    "TaskPythonBuild",
    "ContainerTaskConfig",
    "DatabricksJobTaskConfig",
    "PythonTaskConfig",
    "ExecutionConfig",
    "FlyteFile",
    "FlyteError",
    "PySparkTaskConfig",
]


# Register the SignedURLFileSystem implementation for fsspec
if ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_HOST:
    fsspec.register_implementation("s3", SignedURLFileSystem, clobber=True)
    fsspec.register_implementation("gs", SignedURLFileSystem, clobber=True)
    fsspec.register_implementation("abfs", SignedURLFileSystem, clobber=True)



import sys
import time
from typing import Optional, TypeVar

import socketio
from rich.status import Status

from truefoundry.common.utils import poll_for_function
from truefoundry.deploy._autogen import models as autogen_models
from truefoundry.deploy.builder.docker_service import env_has_docker
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.dao.workspace import get_workspace_by_fqn
from truefoundry.deploy.lib.model.entity import Deployment, DeploymentTransitionStatus
from truefoundry.deploy.lib.util import (
    get_application_fqn_from_deployment_fqn,
    validate_local_source_paths,
)
from truefoundry.deploy.v2.lib.models import BuildResponse
from truefoundry.deploy.v2.lib.source import (
    local_source_to_image,
    local_source_to_remote_source,
)
from truefoundry.logger import logger
from truefoundry.pydantic_v1 import BaseModel

Component = TypeVar("Component", bound=BaseModel)


def _set_python_version_if_missing(build_spec: autogen_models.PythonBuild) -> None:
    if build_spec.python_version is None:
        client = ServiceFoundryServiceClient()
        server_default_python_version = (
            client.python_sdk_config.python_build_default_image_tag
        )
        logger.warning(
            f"No python version was provided in the spec, "
            f"using the default python version ({server_default_python_version}) from the server."
        )
        build_spec.python_version = server_default_python_version


def _handle_if_local_source(component: Component, workspace_fqn: str) -> Component:
    if (
        hasattr(component, "image")
        and isinstance(component.image, autogen_models.Build)
        and isinstance(component.image.build_source, autogen_models.LocalSource)
    ):
        new_component = component.copy(deep=True)
        validate_local_source_paths(
            component_name=new_component.name, build=new_component.image
        )

        if new_component.image.build_source.local_build:
            if not env_has_docker():
                logger.warning(
                    "Did not find Docker locally installed on this system, image will be built remotely. "
                    "For faster builds it is recommended to install Docker locally. "
                    "If you always want to build remotely, "
                    "please set `image.build_source.local_build` to `false` in your YAML spec or equivalently set "
                    "`image=Build(build_source=LocalSource(local_build=False, ...))` in your "
                    "`Service` or `Job` definition code."
                )
                local_build = False
            else:
                logger.info(
                    "Found locally installed docker, image will be built locally and then pushed.\n"
                    "If you want to always build remotely instead of locally, "
                    "please set `image.build_source.local_build` to `false` in your YAML spec or equivalently set "
                    "`image=Build(build_source=LocalSource(local_build=False, ...))` in your "
                    "`Service` or `Job` definition code."
                )
                local_build = True
        else:
            logger.info(
                "Image will be built remotely because `image.build_source.local_build` is set to `false`. "
                "For faster builds it is recommended to install Docker locally and "
                "set `image.build_source.local_build` to `true` in your YAML spec "
                "or equivalently set `image=Build(build_source=LocalSource(local_build=True, ...))` "
                "in your `Service` or `Job` definition code."
            )
            local_build = False

        if local_build:
            if isinstance(new_component.image.build_spec, autogen_models.PythonBuild):
                _set_python_version_if_missing(new_component.image.build_spec)
            # We are to build the image locally, push and update `image` in spec
            logger.info(
                "Building image for %s '%s'", new_component.type, new_component.name
            )
            new_component.image = local_source_to_image(
                build=new_component.image,
                docker_registry_fqn=new_component.image.docker_registry,
                workspace_fqn=workspace_fqn,
                component_name=new_component.name,
            )
        else:
            # We'll build image on TrueFoundry servers, upload the source and update image.build_source
            logger.info(
                "Uploading code for %s '%s'", new_component.type, new_component.name
            )
            client = ServiceFoundryServiceClient()
            new_component.image.build_source = local_source_to_remote_source(
                local_source=new_component.image.build_source,
                workspace_fqn=workspace_fqn,
                component_name=new_component.name,
                upload_code_package=client.upload_code_package,
            )
            logger.debug(
                "Uploaded code for %s '%s'", new_component.type, new_component.name
            )
        return new_component
    return component


def _log_application_dashboard_url(deployment: Deployment, log_message: str):
    application_id = deployment.applicationId

    # TODO: is there any simpler way to get this? :cry
    client = ServiceFoundryServiceClient()

    url = f"{client.tfy_host.strip('/')}/applications/{application_id}?tab=deployments"
    logger.info(log_message, url)


def _tail_build_logs(build_response: BuildResponse) -> socketio.Client:
    client = ServiceFoundryServiceClient()

    logger.info("Tailing build logs for '%s'", build_response.componentName)
    socket = client.tail_build_logs(build_response=build_response)
    return socket


def _deploy_wait_handler(  # noqa: C901
    deployment: Deployment,
    tail_logs: bool = True,
) -> Optional[DeploymentTransitionStatus]:
    tail_logs_or_polling_status_message = (
        "You can press Ctrl + C to exit the tailing of build logs "
    )
    if not tail_logs:
        tail_logs_or_polling_status_message = (
            "you can press Ctrl + C to exit the polling of deployment status "
        )
    _log_application_dashboard_url(
        deployment=deployment,
        log_message=(
            "You can track the progress below or on the dashboard:- '%s'\n"
            f"{tail_logs_or_polling_status_message}"
            "and deployment will continue on the server"
        ),
    )
    with Status(status="Polling for deployment status") as spinner:
        last_status_printed = None
        client = ServiceFoundryServiceClient()
        start_time = time.monotonic()
        total_timeout_time: int = 600
        poll_interval_seconds = 5
        time_elapsed = 0
        socket: socketio.Client = None

        for deployment_statuses in poll_for_function(
            client.get_deployment_statuses,
            poll_after_secs=poll_interval_seconds,
            application_id=deployment.applicationId,
            deployment_id=deployment.id,
            retry_count=10,
        ):
            if len(deployment_statuses) == 0:
                logger.warning("Did not receive any deployment status")
                continue

            latest_deployment_status = deployment_statuses[-1]

            status_to_print = (
                latest_deployment_status.transition or latest_deployment_status.status
            )
            spinner.update(status=f"Current state: {status_to_print}")
            if status_to_print != last_status_printed:
                if DeploymentTransitionStatus.is_failure_state(status_to_print):
                    logger.error("State: %r", status_to_print)
                else:
                    logger.info("State: %r", status_to_print)
                last_status_printed = status_to_print

            if latest_deployment_status.state.isTerminalState:
                if socket and socket.connected:
                    socket.disconnect()
                break

            if (
                latest_deployment_status.transition
                == DeploymentTransitionStatus.BUILDING
            ):
                if tail_logs and not socket:
                    try:
                        build_responses = client.get_deployment_build_response(
                            application_id=deployment.applicationId,
                            deployment_id=deployment.id,
                        )
                        socket = _tail_build_logs(build_responses[0])
                    except Exception as e:
                        logger.error("Error tailing build logs: %s", e)

            time_elapsed = time.monotonic() - start_time
            if time_elapsed > total_timeout_time:
                logger.warning(
                    "Polled build logs for %s secs. Disconnecting from server, the deployment will still continue.",
                    int(time_elapsed),
                )
                if socket and socket.connected:
                    socket.disconnect()
                break

    return last_status_printed


def _resolve_workspace_fqn(
    component: Component, workspace_fqn: Optional[str] = None
) -> str:
    if not workspace_fqn:
        if hasattr(component, "workspace_fqn") and component.workspace_fqn:
            resolved_workspace_fqn = component.workspace_fqn
        else:
            raise ValueError(
                f"""\
No Workspace FQN was provided or mentioned in the spec.
Either add a `workspace_fqn` to your yaml spec as

```
name: {getattr(component, "name", "my-app")}
type: {getattr(component, "type", "undefined")}
...
workspace_fqn: <your workspace fqn>
```

or Python deployment spec as

```
app = {component.__class__.__name__}(
    name='{getattr(component, "name", "my-app")}',
    ...
    workspace_fqn='<your workspace fqn>'
)
```

or pass it explicitly using `--workspace-fqn` argument on CLI.
"""
            )
    else:
        if (
            hasattr(component, "workspace_fqn")
            and component.workspace_fqn
            and component.workspace_fqn != workspace_fqn
        ):
            logger.warning(
                f"`workspace_fqn` set in the deployment spec doesn't match the provided `workspace_fqn` argument {component.workspace_fqn!r} \n"
                f"Using `workspace_fqn`: {workspace_fqn!r} "
            )
        resolved_workspace_fqn = workspace_fqn

    return resolved_workspace_fqn


def deploy_component(
    component: Component,
    workspace_fqn: Optional[str] = None,
    wait: bool = True,
    force: bool = False,
    trigger_on_deploy: bool = False,
) -> Deployment:
    workspace_fqn = _resolve_workspace_fqn(
        component=component, workspace_fqn=workspace_fqn
    )
    component.workspace_fqn = workspace_fqn
    workspace_id = get_workspace_by_fqn(workspace_fqn).id
    if isinstance(component, autogen_models.ApplicationSet):
        updated_component = component.copy(deep=True)
        if updated_component.components:
            for i, subcomponent in enumerate(updated_component.components):
                updated_component.components[i] = _handle_if_local_source(
                    component=subcomponent, workspace_fqn=workspace_fqn
                )
    else:
        updated_component = _handle_if_local_source(
            component=component, workspace_fqn=workspace_fqn
        )
    client = ServiceFoundryServiceClient()
    response = client.deploy_application(
        workspace_id=workspace_id,
        application=updated_component,
        force=force,
        trigger_on_deploy=trigger_on_deploy,
    )
    logger.info(
        "🚀 Deployment started for application '%s'. Deployment FQN is '%s'.",
        updated_component.name,
        response.fqn,
    )
    if wait:
        try:
            last_status_printed = _deploy_wait_handler(deployment=response)
            if not last_status_printed or DeploymentTransitionStatus.is_failure_state(
                last_status_printed
            ):
                deployment_tab_url = f"{client.tfy_host.strip('/')}/applications/{response.applicationId}?tab=deployments"
                message = f"Deployment Failed. Please refer to the logs for additional details - {deployment_tab_url}"
                sys.exit(message)
        except KeyboardInterrupt:
            logger.info("Ctrl-c executed. The deployment will still continue.")

    deployment_fqn = response.fqn
    application_fqn = get_application_fqn_from_deployment_fqn(deployment_fqn)
    logger.info("Deployment FQN: %s", deployment_fqn)
    logger.info("Application FQN: %s", application_fqn)

    _log_application_dashboard_url(
        deployment=response,
        log_message="You can find the application on the dashboard:- '%s'",
    )
    return response



import re

from truefoundry.ml.exceptions import MlFoundryException

KEY_REGEX = re.compile(r"^[a-zA-Z0-9-_]+$")


def validate_key_name(key: str):
    if not key or not KEY_REGEX.match(key):
        raise MlFoundryException(
            f"Invalid run image key: {key} should only contain alphanumeric, hyphen or underscore"
        )



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

import json
import re  # noqa: F401

from aenum import Enum


class LibraryName(str, Enum):
    """
    Name of the library for the framework
    """

    """
    allowed enum values
    """
    TRANSFORMERS = "transformers"
    SENTENCE_MINUS_TRANSFORMERS = "sentence-transformers"
    DIFFUSERS = "diffusers"

    @classmethod
    def from_json(cls, json_str: str) -> LibraryName:
        """Create an instance of LibraryName from a JSON string"""
        return LibraryName(json.loads(json_str))



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.models.metric_dto import MetricDto
from truefoundry.ml._autogen.client.models.param_dto import ParamDto
from truefoundry.ml._autogen.client.models.run_tag_dto import RunTagDto
from truefoundry.pydantic_v1 import BaseModel, StrictStr, conlist


class LogBatchRequestDto(BaseModel):
    """
    LogBatchRequestDto
    """

    run_id: Optional[StrictStr] = None
    metrics: Optional[conlist(MetricDto)] = None
    params: Optional[conlist(ParamDto)] = None
    tags: Optional[conlist(RunTagDto)] = None
    __properties = ["run_id", "metrics", "params", "tags"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> LogBatchRequestDto:
        """Create an instance of LogBatchRequestDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in metrics (list)
        _items = []
        if self.metrics:
            for _item in self.metrics:
                if _item:
                    _items.append(_item.to_dict())
            _dict["metrics"] = _items
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in params (list)
        _items = []
        if self.params:
            for _item in self.params:
                if _item:
                    _items.append(_item.to_dict())
            _dict["params"] = _items
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in tags (list)
        _items = []
        if self.tags:
            for _item in self.tags:
                if _item:
                    _items.append(_item.to_dict())
            _dict["tags"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> LogBatchRequestDto:
        """Create an instance of LogBatchRequestDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return LogBatchRequestDto.parse_obj(obj)

        _obj = LogBatchRequestDto.parse_obj(
            {
                "run_id": obj.get("run_id"),
                "metrics": [MetricDto.from_dict(_item) for _item in obj.get("metrics")]
                if obj.get("metrics") is not None
                else None,
                "params": [ParamDto.from_dict(_item) for _item in obj.get("params")]
                if obj.get("params") is not None
                else None,
                "tags": [RunTagDto.from_dict(_item) for _item in obj.get("tags")]
                if obj.get("tags") is not None
                else None,
            }
        )
        return _obj



import rich_click as click

from truefoundry.cli.const import GROUP_CLS
from truefoundry.ml.cli.commands import download

click.rich_click.USE_RICH_MARKUP = True


@click.group(
    name="ml",
    cls=GROUP_CLS,
)
def ml():
    """
    TrueFoundry ML CLI
    """
    pass


def get_ml_cli():
    ml.add_command(download)
    return ml



import os
import sys
from pathlib import Path
from typing import Dict, List, Optional, Union

import requirements
from flytekit.configuration import (
    SERIALIZED_CONTEXT_ENV_VAR,
    ImageConfig,
    SerializationSettings,
)
from flytekit.configuration import Image as FlytekitImage
from flytekit.models.launch_plan import LaunchPlan as FlyteLaunchPlan
from flytekit.tools.repo import (
    serialize_get_control_plane_entities as get_serialized_entities,
)
from flytekit.tools.repo import serialize_load_only as serialize_workflow
from flytekit.tools.translator import TaskSpec as FlyteTaskSpec
from flytekit.tools.translator import WorkflowSpec as FlyteWorkflowSpec
from google.protobuf.json_format import MessageToDict

from truefoundry.common.types import UploadCodePackageCallable
from truefoundry.deploy._autogen import models as autogen_models
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.dao.workspace import get_workspace_by_fqn
from truefoundry.deploy.lib.model.entity import Deployment, DeploymentTransitionStatus
from truefoundry.deploy.v2.lib.deploy import _deploy_wait_handler
from truefoundry.deploy.v2.lib.source import (
    local_source_to_remote_source,
)
from truefoundry.logger import logger
from truefoundry.pydantic_v1 import ValidationError
from truefoundry.workflow.workflow import (
    TRUEFOUNDRY_ALERTS_CONFIG,
    TRUEFOUNDRY_LAUNCH_PLAN_NAME,
    truefoundry_config_store,
)


def _handle_code_upload_for_workflow(
    workflow: autogen_models.Workflow,
    workspace_fqn: str,
    upload_code_package: UploadCodePackageCallable,
) -> autogen_models.Workflow:
    new_workflow = workflow.copy(deep=True)
    new_workflow.source = local_source_to_remote_source(
        local_source=workflow.source,
        workspace_fqn=workspace_fqn,
        component_name=workflow.name,
        upload_code_package=upload_code_package,
    )
    return new_workflow


def _is_tfy_workflow_package(package: requirements.parser.Requirement) -> bool:
    return package.name == "truefoundry" and "workflow" in package.extras


def _is_tfy_wf_present_in_pip_packages_or_requirements_file(
    pip_packages: List[str],
    project_root_path: str,
    requirements_path: Optional[str] = None,
) -> bool:
    for package in pip_packages:
        parsed_package = requirements.parser.Requirement.parse(package)
        if _is_tfy_workflow_package(parsed_package):
            return True
    if requirements_path:
        requirements_file_absolute_path = os.path.join(
            project_root_path, requirements_path
        )
        if not os.path.exists(requirements_file_absolute_path):
            raise FileNotFoundError(
                f"requirements file not found at {requirements_file_absolute_path}. requirements file path should be relative to project root path."
            )
        with open(requirements_file_absolute_path, "r") as file:
            for package in requirements.parse(file):
                if _is_tfy_workflow_package(package):
                    return True
    return False


def _is_tfy_wf_present_in_task_python_build(
    task_image_spec: Dict, project_root_path: str
) -> bool:
    pip_packages = task_image_spec["pip_packages"] or []
    requirements_path = task_image_spec.get("requirements_path")
    return _is_tfy_wf_present_in_pip_packages_or_requirements_file(
        pip_packages=pip_packages,
        project_root_path=project_root_path,
        requirements_path=requirements_path,
    )


def _is_dynamic_task(flyte_task: FlyteTaskSpec) -> bool:
    envs: Dict[str, str] = flyte_task.template.container.env or {}
    return SERIALIZED_CONTEXT_ENV_VAR in envs.keys()


# this function does validation that num_workflows = 1, this also validates task_config is passed correctly.
# This is verified by pydantic but doing it here also as error messages are not clear in pydantic
def _validate_workflow_entities(  # noqa: C901
    workflow_entities: List[Union[FlyteWorkflowSpec, FlyteLaunchPlan, FlyteTaskSpec]],
    project_root_path: str,
):
    workflow_objs: List[FlyteWorkflowSpec] = []
    launch_plans: List[FlyteLaunchPlan] = []
    tasks: List[FlyteTaskSpec] = []
    for entity in workflow_entities:
        if isinstance(entity, FlyteWorkflowSpec):
            workflow_objs.append(entity)
        elif isinstance(entity, FlyteLaunchPlan):
            launch_plans.append(entity)
        elif isinstance(entity, FlyteTaskSpec):
            tasks.append(entity)
        else:
            raise ValueError(f"Invalid entity found in workflow: {entity}")
    if len(workflow_objs) != 1:
        raise ValueError(
            f"Workflow file must have exactly one workflow object. Found {len(workflow_objs)}"
        )
    if len(launch_plans) > 2:
        raise ValueError(
            f"Workflow file must have exactly one launch plan. Found {len(launch_plans) - 1}"
        )

    error_message_to_use_truefoundry_decorators = """Invalid task definition for task: {}, Please use valid truefoundry decorator/class and pass task_config for tasks.
        You can import truefoundry task decorators using:
        `from truefoundry.workflow import task, ContainerTask, map_task`
        You can pass task config using `task_config` parameter in the task definition. Task config should be one of the following:
        `PythonTaskConfig`, `ContainerTaskConfig`, `PySparkTaskConfig`, or `DatabricksJobTaskConfig`. You can import these using:
        `from truefoundry.workflow import PythonTaskConfig, ContainerTaskConfig, PySparkTaskConfig, DatabricksJobTaskConfig`
         """
    tasks_without_truefoundry_worflow_package = []
    task_names = set()
    for task in tasks:
        if _is_dynamic_task(task):
            raise ValueError("Dynamic workflows are not supported yet.")
        if not task.template.custom:
            raise ValueError(
                error_message_to_use_truefoundry_decorators.format(
                    task.template.id.name
                )
            )
        task_name = task.template.id.name
        if task_name in task_names:
            raise ValueError(
                f"Task name should be unique, task with name {task_name} is repeated in the workflow"
            )
        task_names.add(task_name)
        task_custom = task.template.custom["truefoundry"]
        task_image_spec = task_custom.get("image")
        if task_image_spec and task_image_spec.get("type") == "task-python-build":
            is_tfy_wf_present_in_task_python_build = (
                _is_tfy_wf_present_in_task_python_build(
                    task_image_spec=task_image_spec, project_root_path=project_root_path
                )
            )
            if not is_tfy_wf_present_in_task_python_build:
                tasks_without_truefoundry_worflow_package.append(task.template.id.name)
        try:
            autogen_models.FlyteTaskCustom.validate(task.template.custom)
        except ValidationError:
            raise ValueError(
                error_message_to_use_truefoundry_decorators.format(
                    task.template.id.name
                )
            ) from None
    if len(tasks_without_truefoundry_worflow_package) > 0:
        raise ValueError(
            rf"truefoundry\[workflow] package is required dependency to run workflows, add it in pip_packages for tasks: {', '.join(tasks_without_truefoundry_worflow_package)}"
        )

    # validate that all inputs have default values for cron workflows
    for launch_plan in launch_plans:
        if (
            truefoundry_config_store.get(launch_plan.spec.workflow_id.name)
            and launch_plan.id.name == TRUEFOUNDRY_LAUNCH_PLAN_NAME
        ):
            workflow_inputs = launch_plan.spec.default_inputs.parameters
            for input in workflow_inputs:
                if workflow_inputs[input].required:
                    raise ValueError(
                        f"All inputs must have default for a cron workflow. Input {input} is required in workflow but default value is not provided"
                    )


def _get_relative_package_path_from_filepath(
    project_root_path: str, filepath: str
) -> str:
    """
    This function returns the relative package path from the project root path for a given file path.
    e.g. if project_root_path = /home/user/project and filepath = /home/user/project/src/module.py
    then the function will return src.module
    """
    relative_file_path = os.path.relpath(filepath, project_root_path)
    path = Path(relative_file_path)
    package_path = str(path.with_suffix("")).replace(os.path.sep, ".")

    return package_path


def _generate_manifest_for_workflow(
    workflow: autogen_models.Workflow,
):
    settings = SerializationSettings(
        # We are adding these defaults to avoid validation errors in flyte objects
        image_config=ImageConfig(default_image=FlytekitImage(name="", tag="", fqn="")),
        python_interpreter=sys.executable,
    )
    source_absolute_path = os.path.abspath(workflow.source.project_root_path)

    workflow_file_absolute_path = os.path.join(
        source_absolute_path, workflow.workflow_file_path
    )
    if not os.path.exists(workflow_file_absolute_path):
        raise FileNotFoundError(
            f"Workflow file not found at {workflow_file_absolute_path}. Workflow file path should be relative to project root path."
        )

    package_path = _get_relative_package_path_from_filepath(
        project_root_path=source_absolute_path, filepath=workflow_file_absolute_path
    )

    serialize_workflow(
        pkgs=[package_path], settings=settings, local_source_root=source_absolute_path
    )
    workflow_entities = get_serialized_entities(
        settings, local_source_root=source_absolute_path
    )
    _validate_workflow_entities(workflow_entities, source_absolute_path)

    workflow.flyte_entities = []
    for entity in workflow_entities:
        if isinstance(entity, FlyteLaunchPlan):
            workflow_name = entity.spec.workflow_id.name

            # this is the case when someone has a cron schedule. and this line is for handling default launch plan in this case.
            if (
                truefoundry_config_store.get(workflow_name)
                and workflow_name == entity.id.name
            ):
                continue
            # this is the case when someone does not have a cron schedule. and this line is for handling default launch plan in this case.
            elif entity.id.name == workflow_name:
                entity._id._name = TRUEFOUNDRY_LAUNCH_PLAN_NAME
            # this the case when some workflow doesn't have cron schedule, neither it is default launch plan
            elif entity.id.name != TRUEFOUNDRY_LAUNCH_PLAN_NAME:
                raise ValueError(
                    f"Creating launch plans is not allowed. Found launch plan with name {entity.id.name}"
                )
        message_dict = MessageToDict(entity.to_flyte_idl())
        # proto message to dict conversion converts all int to float. so we need this hack
        if (
            message_dict.get("template")
            and message_dict["template"].get("custom")
            and message_dict["template"]["custom"].get("truefoundry")
        ):
            parsed_model = autogen_models.FlyteTaskCustom.parse_obj(
                message_dict["template"]["custom"]
            )
            message_dict["template"]["custom"]["truefoundry"] = parsed_model.truefoundry

        workflow.flyte_entities.append(message_dict)

    alerts_config = truefoundry_config_store.get(TRUEFOUNDRY_ALERTS_CONFIG)
    if alerts_config:
        if not workflow.alerts:
            workflow.alerts = alerts_config
        else:
            logger.warning(
                "Alerts are configured in both workflow decorator as well as in deployment config. Alerts configured in workflow decorator will be ignored."
            )
    # this step is just to verify if pydantic model is still valid after adding flyte_entities
    autogen_models.Workflow.validate({**workflow.dict()})


def _validate_workspace_fqn(
    workflow: autogen_models.Workflow,
    workspace_fqn: Optional[str] = None,
):
    if not workspace_fqn:
        raise ValueError(
            "No Workspace FQN was provided. "
            "Pass it explicitly using `--workspace-fqn` argument on CLI or `workspace_fqn` argument of `deploy_workflow`."
        )


def deploy_workflow(
    workflow: autogen_models.Workflow,
    workspace_fqn: str,
    wait: bool = True,
    force: bool = False,
) -> Deployment:
    _generate_manifest_for_workflow(workflow)
    _validate_workspace_fqn(workflow, workspace_fqn)

    # we need to rest the execution config store as it is a global variable and we don't want to keep the cron execution config for next workflow
    # this is only needed for notebook environment
    truefoundry_config_store.reset()

    workspace_id = get_workspace_by_fqn(workspace_fqn).id

    logger.info(
        f"Deploying workflow {workflow.name} to workspace {workspace_fqn} ({workspace_id})"
    )
    client = ServiceFoundryServiceClient()

    workflow = _handle_code_upload_for_workflow(
        workflow=workflow,
        workspace_fqn=workspace_fqn,
        upload_code_package=client.upload_code_package,
    )

    deployment = client.deploy_application(
        workspace_id=workspace_id,
        application=workflow,
        force=force,
    )
    logger.info(
        "🚀 Deployment started for application '%s'. Deployment FQN is '%s'.",
        workflow.name,
        deployment.fqn,
    )
    deployment_url = f"{client.tfy_host.strip('/')}/applications/{deployment.applicationId}?tab=deployments"
    if wait:
        try:
            last_status_printed = _deploy_wait_handler(
                deployment=deployment, tail_logs=False
            )
            if not last_status_printed or DeploymentTransitionStatus.is_failure_state(
                last_status_printed
            ):
                deployment_tab_url = f"{client.tfy_host.strip('/')}/applications/{deployment.applicationId}?tab=deployments"
                message = f"Deployment Failed. Please refer to the logs for additional details - {deployment_tab_url}"
                sys.exit(message)
        except KeyboardInterrupt:
            logger.info("Ctrl-c executed. The deployment will still continue.")
    logger.info("You can find the application on the dashboard:- '%s'", deployment_url)



import argparse
import importlib
import json
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import numpy as np

from truefoundry.ml.exceptions import MlFoundryException


def get_module(
    module_name: str, error_message: Optional[str] = None, required: bool = False
):
    try:
        return importlib.import_module(module_name)
    except Exception as ex:
        msg = error_message or f"Error importing module {module_name}"
        if required:
            raise MlFoundryException(msg) from ex


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


ParamsType = Union[Mapping[str, Any], argparse.Namespace]


def process_params(params: ParamsType) -> Mapping[str, Any]:
    if isinstance(params, Mapping):
        return params
    if isinstance(params, argparse.Namespace):
        return vars(params)
    # TODO: add absl support if required
    # move to a different file then
    raise MlFoundryException(
        "params should be either argparse.Namespace or a Mapping (dict) type"
    )


def flatten_dict(
    input_dict: Mapping[Any, Any], parent_key: str = "", sep: str = "."
) -> Dict[str, Any]:
    """Flatten hierarchical dict, e.g. ``{'a': {'b': 'c'}} -> {'a.b': 'c'}``.
    All the keys will be converted to str.

    Args:
        input_dict: Dictionary containing the keys
        parent_key: Prefix to add to the keys. Defaults to ``''``.
        sep: Delimiter to express the hierarchy. Defaults to ``'.'``.

    Returns:
        Flattened dict.

    Examples:
        >>> flatten_dict({'a': {'b': 'c'}})
        {'a.b': 'c'}
        >>> flatten_dict({'a': {'b': 123}})
        {'a.b': 123}
        >>> flatten_dict({'a': {'b': 'c'}}, parent_key="param")
        {'param.a.b': 'c'}
    """
    new_dict_items: List[Tuple[str, Any]] = []
    for k, v in input_dict.items():
        new_key = parent_key + sep + str(k) if parent_key else k
        if isinstance(v, Mapping):
            new_dict_items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            new_dict_items.append((new_key, v))
    return dict(new_dict_items)



import json
import logging
import os
import platform
import sys
import time
from collections import namedtuple
from functools import lru_cache, wraps
from shutil import rmtree
from subprocess import check_output
from time import monotonic_ns
from typing import Callable, Generator, List, Optional, TypeVar
from urllib.parse import urljoin, urlparse

from truefoundry.common.constants import (
    API_SERVER_RELATIVE_PATH,
    ENV_VARS,
    MLFOUNDRY_SERVER_RELATIVE_PATH,
    TFY_API_KEY_ENV_KEY,
    TFY_DEBUG_ENV_KEY,
    TFY_HOST_ENV_KEY,
    TFY_INTERNAL_ENV_KEY,
)
from truefoundry.pydantic_v1 import BaseSettings
from truefoundry.version import __version__

logger = logging.getLogger(__name__)

T = TypeVar("T")
InstalledPipPackage = namedtuple("InstalledPipPackage", ["name", "version"])


class _TFYServersConfig(BaseSettings):
    class Config:
        env_prefix = "TFY_CLI_LOCAL_"
        env_file = ".tfy-cli-local.env"

    tenant_host: str
    servicefoundry_server_url: str
    mlfoundry_server_url: str

    @classmethod
    def from_tfy_host(cls, tfy_host: str) -> "_TFYServersConfig":
        tfy_host = tfy_host.strip("/")
        return cls(
            tenant_host=urlparse(tfy_host).netloc,
            servicefoundry_server_url=urljoin(tfy_host, API_SERVER_RELATIVE_PATH),
            mlfoundry_server_url=urljoin(tfy_host, MLFOUNDRY_SERVER_RELATIVE_PATH),
        )


_tfy_servers_config = None


def get_tfy_servers_config(tfy_host: str) -> _TFYServersConfig:
    global _tfy_servers_config
    if _tfy_servers_config is None:
        if ENV_VARS.TFY_CLI_LOCAL_DEV_MODE:
            _tfy_servers_config = _TFYServersConfig()
        else:
            _tfy_servers_config = _TFYServersConfig.from_tfy_host(tfy_host)
    return _tfy_servers_config


def relogin_error_message(message: str, host: str = "HOST") -> str:
    suffix = ""
    if host == "HOST":
        suffix = " where HOST is TrueFoundry platform URL"
    return (
        f"{message}\n"
        f"To relogin, use `tfy login --host {host} --relogin` "
        f"or `truefoundry.login(host={host!r}, relogin=True)` function" + suffix
    )


def timed_lru_cache(
    seconds: int = 300, maxsize: Optional[int] = None
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def wrapper_cache(func: Callable[..., T]) -> Callable[..., T]:
        func = lru_cache(maxsize=maxsize)(func)
        func.delta = seconds * 10**9
        func.expiration = monotonic_ns() + func.delta

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if monotonic_ns() >= func.expiration:
                func.cache_clear()
                func.expiration = monotonic_ns() + func.delta
            return func(*args, **kwargs)

        return wrapped_func

    return wrapper_cache


def poll_for_function(
    func: Callable[..., T],
    poll_after_secs: int = 5,
    *args,
    **kwargs,
) -> Generator[T, None, None]:
    while True:
        yield func(*args, **kwargs)
        time.sleep(poll_after_secs)


def validate_tfy_host(tfy_host: str) -> None:
    if not (tfy_host.startswith("https://") or tfy_host.startswith("http://")):
        raise ValueError(
            f"Invalid host {tfy_host!r}. It should start with https:// or http://"
        )


def resolve_tfy_host(tfy_host: Optional[str] = None) -> str:
    tfy_host = tfy_host or ENV_VARS.TFY_HOST
    if not tfy_host:
        if ENV_VARS.TFY_API_KEY:
            raise ValueError(
                f"TFY_HOST` env must be set since `{TFY_API_KEY_ENV_KEY}` env is set. Either set `{TFY_HOST_ENV_KEY}` or unset `{TFY_API_KEY_ENV_KEY}` and login"
            )
        else:
            raise ValueError(
                f"Either `host` should be provided using `--host <value>`, or `{TFY_HOST_ENV_KEY}` env must be set"
            )
    tfy_host = tfy_host.strip("/")
    validate_tfy_host(tfy_host)
    return tfy_host


class ContextualDirectoryManager:
    def __init__(self, dir_path: str, cleanup_on_error: bool = True):
        self.dir_path = dir_path
        self.cleanup_on_error = cleanup_on_error

    def __enter__(self):
        if os.path.exists(self.dir_path):
            raise FileExistsError(
                f"The directory {self.dir_path!r} already exists. "
                "Please provide a path with a different name that does not already exist."
            )

        os.makedirs(self.dir_path, exist_ok=False)
        return self.dir_path

    def __exit__(self, exc_type, exc_value, traceback):
        if self.cleanup_on_error and exc_type is not None:
            # Only delete the directory if an exception occurred
            if os.path.exists(self.dir_path):
                rmtree(self.dir_path)


def get_python_version_major_minor() -> str:
    """
    Returns the major.minor version of the Python interpreter
    """
    version_info = sys.version_info
    return f"{version_info.major}.{version_info.minor}"


def list_pip_packages_installed(
    filter_package_names: Optional[List[str]] = None,
) -> List[InstalledPipPackage]:
    """
    List the installed package_names, along with their versions.
    Args:
        filter_package_names (List[str]): A list of specific libraries to filter for.

    Returns:
        List[InstalledPipPackage]: A list of InstalledPipPackage namedtuples for each match.
    """
    relevant_package_names: List[InstalledPipPackage] = []

    # Get the installed packages in JSON format from pip
    try:
        output = check_output(
            [
                sys.executable,
                "-m",
                "pip",
                "--disable-pip-version-check",
                "list",
                "--pre",
                "--format=json",
            ]
        )
        installed_package_names = json.loads(output.decode("utf-8"))
    except Exception:
        logger.exception("Failed to list installed packages using pip.")
        return relevant_package_names

    package_names_to_check = set(filter_package_names or [])
    for package in installed_package_names:
        if package["name"] in package_names_to_check:
            relevant_package_names.append(
                InstalledPipPackage(package["name"], package["version"])
            )
    return relevant_package_names


def is_debug_env_set() -> bool:
    return (os.getenv(TFY_DEBUG_ENV_KEY) or "").lower() in ["true", "1"]


def is_internal_env_set() -> bool:
    return (os.getenv(TFY_INTERNAL_ENV_KEY) or "").lower() in ["true", "1"]


def get_user_agent() -> str:
    try:
        return f"truefoundry/{__version__} Python/{platform.python_version()} OS/{platform.system()}-{platform.release()} ({platform.architecture()[0]})"
    except Exception:
        return f"truefoundry/{__version__}"


def get_expanded_and_absolute_path(path: str):
    return os.path.abspath(os.path.expanduser(path))



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.models.multi_part_upload_storage_provider import (
    MultiPartUploadStorageProvider,
)
from truefoundry.ml._autogen.client.models.signed_url_dto import SignedURLDto
from truefoundry.pydantic_v1 import BaseModel, Field, StrictStr, conlist


class MultiPartUploadDto(BaseModel):
    """
    MultiPartUploadDto
    """

    storage_provider: MultiPartUploadStorageProvider = Field(...)
    part_signed_urls: conlist(SignedURLDto) = Field(...)
    s3_compatible_upload_id: Optional[StrictStr] = None
    azure_blob_block_ids: Optional[conlist(StrictStr)] = None
    finalize_signed_url: SignedURLDto = Field(...)
    __properties = [
        "storage_provider",
        "part_signed_urls",
        "s3_compatible_upload_id",
        "azure_blob_block_ids",
        "finalize_signed_url",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> MultiPartUploadDto:
        """Create an instance of MultiPartUploadDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in part_signed_urls (list)
        _items = []
        if self.part_signed_urls:
            for _item in self.part_signed_urls:
                if _item:
                    _items.append(_item.to_dict())
            _dict["part_signed_urls"] = _items
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of finalize_signed_url
        if self.finalize_signed_url:
            _dict["finalize_signed_url"] = self.finalize_signed_url.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> MultiPartUploadDto:
        """Create an instance of MultiPartUploadDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return MultiPartUploadDto.parse_obj(obj)

        _obj = MultiPartUploadDto.parse_obj(
            {
                "storage_provider": obj.get("storage_provider"),
                "part_signed_urls": [
                    SignedURLDto.from_dict(_item)
                    for _item in obj.get("part_signed_urls")
                ]
                if obj.get("part_signed_urls") is not None
                else None,
                "s3_compatible_upload_id": obj.get("s3_compatible_upload_id"),
                "azure_blob_block_ids": obj.get("azure_blob_block_ids"),
                "finalize_signed_url": SignedURLDto.from_dict(
                    obj.get("finalize_signed_url")
                )
                if obj.get("finalize_signed_url") is not None
                else None,
            }
        )
        return _obj



from typing import Optional, Union

from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    FileInfoDto,
    MetricDto,
)
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)


class Metric(BaseModel):
    key: StrictStr = Field(...)
    value: Optional[Union[StrictFloat, StrictInt]] = None
    timestamp: Optional[StrictInt] = None
    step: Optional[StrictInt] = 0

    @classmethod
    def from_dto(cls, dto: MetricDto) -> "Metric":
        return cls(
            key=dto.key,
            value=dto.value,
            timestamp=dto.timestamp,
            step=dto.step,
        )

    def to_dto(self) -> MetricDto:
        return MetricDto(
            key=self.key,
            value=self.value,
            timestamp=self.timestamp,
            step=self.step,
        )


class FileInfo(BaseModel):
    path: StrictStr
    is_dir: StrictBool
    file_size: Optional[StrictInt] = None
    signed_url: Optional[StrictStr] = None

    @classmethod
    def from_dto(cls, dto: FileInfoDto) -> "FileInfo":
        return cls(
            path=dto.path,
            is_dir=dto.is_dir,
            file_size=dto.file_size,
            signed_url=dto.signed_url,
        )

    def to_dto(self) -> FileInfoDto:
        return FileInfoDto(
            path=self.path,
            is_dir=self.is_dir,
            file_size=self.file_size,
            signed_url=self.signed_url,
        )



from __future__ import annotations

from typing import Iterable

from rich.console import Console

from truefoundry.autodeploy.tools.base import Event, RequestEvent, ResponseEvent, Tool
from truefoundry.autodeploy.utils.pydantic_compat import model_dump


class AskQuestion(Event):
    question: str

    def render(self, console: Console) -> str:
        console.print(f"[bold magenta]TrueFoundry:[/] {self.question}")
        response = console.input("[bold green]You:[/] ")
        return response


class Ask(Tool):
    description = """
Ask a question to the user.
"""

    class Request(RequestEvent):
        question: str

    class Response(ResponseEvent):
        response: str

    def run(self, request: Ask.Request) -> Iterable[Event]:
        response = yield AskQuestion(**model_dump(request))
        return Ask.Response(response=response)



from __future__ import annotations

import os
from collections import Counter
from typing import Any, Dict, Generator

import gitignorefile
from rich.console import Console
from rich.table import Table

from truefoundry.autodeploy.tools.base import (
    Event,
    Message,
    RequestEvent,
    ResponseEvent,
    Rule,
    Tool,
)
from truefoundry.pydantic_v1 import Field


class ShowFileCount(Event):
    file_types: Dict[str, int]

    def render(self, console: Console):
        total_files = sum(self.file_types.values())
        console.print(f"Found {total_files} files.")
        table = Table()
        table.add_column("File Type", style="cyan")
        table.add_column("Count", justify="right", style="green")
        for file_type, count in self.file_types.items():
            table.add_row(file_type, str(count))

        console.print(table)


class FileTypeCounts(Tool):
    description = """
Get counts of different types of file present.
"""

    class Request(RequestEvent): ...

    class Response(ResponseEvent):
        file_types: Dict[str, int] = Field(
            ...,
            description='Counts of different types of files. Ex: {"py": 1} or {"c": 1}',
        )

    def __init__(self, project_root_path: str):
        self.project_root_path = project_root_path

    def run(
        self, request: FileTypeCounts.Request
    ) -> Generator[Event, Any, ResponseEvent]:
        counter = Counter()
        yield Rule(message="[bold green]File Counts[/]")
        yield Message(
            message="[bold cyan]Processing:[/] Scanning for various file types..."
        )

        def gitignore(_):
            return False

        gitignore_path = os.path.join(self.project_root_path, ".gitignore")
        if os.path.exists(gitignore_path):
            gitignore = gitignorefile.parse(path=gitignore_path)
        for root, dirs, ps in os.walk(
            self.project_root_path,
        ):
            root = root[len(self.project_root_path) :]
            if ".git" in dirs:
                dirs.remove(".git")
            counter.update(
                p.split(".")[-1] if len(p) and p[0] != "." else p
                for p in ps
                if not gitignore(os.path.join(root, p).strip(os.path.sep))
            )
        yield ShowFileCount(file_types=counter)
        return FileTypeCounts.Response(file_types=counter)



"""
Flyte Python task that triggers a Databricks job via the Jobs API.

Authentication uses the Databricks SDK OIDC flow: KubernetesIdTokenSource supplies
the Kubernetes service account token (from /var/run/secrets/kubernetes.io/serviceaccount/token),
and the SDK exchanges it for a Databricks access token via the workspace OIDC endpoint.
The task uses the kubernetes_oidc_strategy credentials strategy with Config.

Execution: always calls jobs.run_now() with idempotency_token derived from the Flyte
execution task ID. If wait_for_completion is true, then waits via
wait_get_run_job_terminated_or_skipped(); on timeout the run is canceled and an
error is raised. If wait_for_completion is false, returns immediately after triggering.

Authentication: set DATABRICKS_PERSONAL_ACCESS_TOKEN for token-based auth, or
DATABRICKS_SERVICE_PRINCIPAL_CLIENT_ID for OIDC (K8s service account token exchange).
The Databricks workspace must have OIDC federation configured when using OIDC (e.g. correct audience).
"""

import logging as logger
import os
from datetime import timedelta
from pathlib import Path
from typing import Any, Callable, Dict, Union

from databricks.sdk import WorkspaceClient, oidc
from databricks.sdk.core import (
    Config,
    CredentialsProvider,
    credentials_strategy,
    oidc_credentials_provider,
)
from databricks.sdk.service.jobs import TerminationCodeCode
from flytekit import PythonFunctionTask, current_context
from flytekit.configuration import SerializationSettings
from flytekit.extend import TaskPlugins
from flytekit.extend.backend.base_agent import AsyncAgentExecutorMixin

from truefoundry.deploy.v2.lib.patched_models import DatabricksJobTaskConfig
from truefoundry.workflow.constants import (
    IDEMPOTENCY_TOKEN_ALLOWED_CHARS,
    IDEMPOTENCY_TOKEN_MAX_LEN,
    SERVICE_ACCOUNT_TOKEN_PATH,
)


def _idempotency_token(execution_id: str) -> str:
    """Build a run-now idempotency token from the Flyte execution ID (max 64 safe chars)."""
    raw = str(execution_id)
    safe = "".join(
        c if c in IDEMPOTENCY_TOKEN_ALLOWED_CHARS else "_" for c in raw
    ).strip("_")
    if len(safe) > IDEMPOTENCY_TOKEN_MAX_LEN:
        safe = safe[:IDEMPOTENCY_TOKEN_MAX_LEN]
        logger.debug(
            "Idempotency token truncated from %s chars to %s",
            len(raw),
            IDEMPOTENCY_TOKEN_MAX_LEN,
        )
    return safe


class KubernetesIdTokenSource(oidc.IdTokenSource):
    """
    IdTokenSource that reads the Kubernetes service account JWT from the default
    mount path (or a custom path) for use in Databricks OIDC token exchange.
    """

    def __init__(
        self,
        client_id: str,
        token_path: Union[str, Path] = SERVICE_ACCOUNT_TOKEN_PATH,
    ):
        self.client_id = client_id
        self.token_path = (
            Path(token_path) if not isinstance(token_path, Path) else token_path
        )

    def id_token(self) -> oidc.IdToken:
        if not self.token_path.exists():
            raise FileNotFoundError(
                f"Service account token not found at {self.token_path}. "
                "Ensure the task pod runs with a service account and the token is mounted."
            )
        token = self.token_path.read_text().strip()
        return oidc.IdToken(jwt=token)


@credentials_strategy("kubernetes-oidc", [])
def kubernetes_oidc_strategy(cfg: Config) -> CredentialsProvider:
    """Credentials strategy that uses the K8s service account token for OIDC exchange."""
    id_source = KubernetesIdTokenSource(client_id=cfg.client_id)
    return oidc_credentials_provider(cfg, id_source)


class TfyDatabricksFunctionTask(
    AsyncAgentExecutorMixin, PythonFunctionTask[DatabricksJobTaskConfig]
):
    """
    Flyte plugin that runs a Databricks job via the Databricks SDK.

    Authenticates using the kubernetes-oidc credentials strategy (K8s service account
    token exchanged for a Databricks access token). Triggers the job with run_now()
    using an idempotency token derived from the Flyte execution task ID. When
    wait_for_completion is true, waits with wait_get_run_job_terminated_or_skipped();
    on timeout the run is canceled and a RuntimeError is raised.
    """

    _DATABRICKS_TASK_TYPE = "python-task"

    def __init__(
        self,
        task_config: DatabricksJobTaskConfig,
        task_function: Callable,
        **kwargs,
    ):
        super().__init__(
            task_config=task_config,
            task_type=self._DATABRICKS_TASK_TYPE,
            task_function=task_function,
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        return {"truefoundry": self._task_config.dict()}

    def execute(self, **kwargs) -> Any:
        cfg = self._task_config
        logger.info(
            "Databricks task execute: workspace_host=%s job_id=%s",
            cfg.workspace_host,
            cfg.job_id,
        )
        ctx = current_context()
        execution_id = ctx.execution_id.name

        idem_token = _idempotency_token(execution_id)
        logger.debug(
            "Flyte execution_id=%s idempotency_token=%s", execution_id, idem_token
        )

        # Get the personal access token from the environment variable
        # If not found, default to using OAuth token federation
        databricks_personal_access_token = os.environ.get(
            "DATABRICKS_PERSONAL_ACCESS_TOKEN"
        )
        if databricks_personal_access_token:
            workspace_client = WorkspaceClient(
                host=cfg.workspace_host,
                token=databricks_personal_access_token,
            )
        elif oidc_client_id := os.environ.get("DATABRICKS_SERVICE_PRINCIPAL_CLIENT_ID"):
            config = Config(
                host=cfg.workspace_host,
                client_id=oidc_client_id,
                credentials_strategy=kubernetes_oidc_strategy,
            )
            workspace_client = WorkspaceClient(config=config)
        else:
            raise RuntimeError(
                "DATABRICKS_SERVICE_PRINCIPAL_CLIENT_ID or DATABRICKS_PERSONAL_ACCESS_TOKEN environment variable is required for using databricks task"
            )
        run = workspace_client.jobs.run_now(
            job_id=int(cfg.job_id),
            idempotency_token=idem_token,
            job_parameters=cfg.job_parameters,
        )

        if not cfg.skip_wait_for_completion:
            timeout = (
                timedelta(seconds=cfg.timeout_seconds)
                if cfg.timeout_seconds is not None
                else timedelta(minutes=20)
            )
            try:
                poll = workspace_client.jobs.wait_get_run_job_terminated_or_skipped(
                    run_id=run.run_id,
                    timeout=timeout,
                )
                if poll.status.termination_details.code != TerminationCodeCode.SUCCESS:
                    logger.error("Databricks job run failed: run_id=%s", run.run_id)
                    raise RuntimeError(
                        f"Databricks job run failed: run_id={run.run_id}"
                    )
                logger.info("Databricks job run completed: run_id=%s", run.run_id)
            except TimeoutError as e:
                workspace_client.jobs.cancel_run(run_id=run.run_id)
                logger.warning(
                    "Canceled Databricks run_id=%s after timeout", run.run_id
                )
                logger.error("Databricks job run timed out: %s", e)
                raise RuntimeError(f"Databricks job run timed out: {e}") from e
        else:
            logger.info(
                "Databricks job triggered (wait_for_completion=false): run_id=%s",
                run.run_id,
            )
        return PythonFunctionTask.execute(self, **kwargs)


TaskPlugins.register_pythontask_plugin(
    DatabricksJobTaskConfig, TfyDatabricksFunctionTask
)



# file: client.py
import io
import os
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlencode, urljoin

from requests.exceptions import RequestException

from truefoundry.common.constants import (
    ENV_VARS,
    TFY_INTERNAL_SIGNED_URL_SERVER_HOST_ENV_KEY,
    TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN_ENV_KEY,
)
from truefoundry.common.request_utils import requests_retry_session
from truefoundry.common.storage_provider_utils import MultiPartUploadStorageProvider
from truefoundry.pydantic_v1 import BaseModel, Field
from truefoundry.workflow.remote_filesystem.logger import log_time, logger

LOG_PREFIX = "[tfy][fs]"
DEFAULT_TTL = ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_DEFAULT_TTL
MAX_TIMEOUT = ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_MAX_TIMEOUT
REQUEST_TIMEOUT = ENV_VARS.TFY_INTERNAL_SIGNED_URL_REQUEST_TIMEOUT
MULTIPART_UPLOAD_FINALIZE_SIGNED_URL_TIMEOUT = (
    ENV_VARS.TFY_INTERNAL_MULTIPART_UPLOAD_FINALIZE_SIGNED_URL_TIMEOUT
)


class SignedURLAPIResponseDto(BaseModel):
    uri: str
    signed_url: str
    headers: Optional[Dict[str, Any]] = None


class SignedURLIsDirectoryAPIResponseDto(BaseModel):
    is_directory: bool = Field(..., alias="isDirectory")


class SignedURLExistsAPIResponseDto(BaseModel):
    exists: bool


class PartSignedUrl(BaseModel):
    partNumber: int
    signedUrl: str


class SignedURLMultipartUploadAPIResponseDto(BaseModel):
    uploadId: str
    partSignedUrls: List[PartSignedUrl]
    finalizeSignedUrl: str
    storageProvider: MultiPartUploadStorageProvider


class FileInfo(BaseModel):
    path: str
    is_directory: bool = Field(..., alias="isDirectory")
    bytes: Optional[int] = None
    signed_url: Optional[str] = None


class PagedList(BaseModel):
    items: List[FileInfo]
    token: Optional[str] = None


class SignedURLServerEndpoint(str, Enum):
    """Enumeration for Signed URL Server endpoints."""

    READ = "/v1/signed-uri/read"
    WRITE = "/v1/signed-uri/write"
    EXISTS = "/v1/exists"
    IS_DIRECTORY = "/v1/is-dir"
    LIST_FILES = "/v1/list-files"
    CREATE_MUTLIPART_UPLOAD = "/v1/multipart-upload"


class SignedURLClient:
    """Client to interact with the Signed URL Server for file operations."""

    def __init__(
        self,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
        ttl: int = DEFAULT_TTL,
        max_retries: int = 3,
        retry_backoff_factor: float = 0.3,
    ):
        """Initialize the SignedURLClient."""
        # Set the base URL and token from the environment variables
        self.base_url = base_url or ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_HOST
        self.token = token or ENV_VARS.TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN
        self.signed_url_server_headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

        if not self.base_url or not self.token:
            raise ValueError(
                f"Set `base_url` and `token` or define env vars "
                f"{TFY_INTERNAL_SIGNED_URL_SERVER_HOST_ENV_KEY} and "
                f"{TFY_INTERNAL_SIGNED_URL_SERVER_TOKEN_ENV_KEY}"
            )

        # Set the TTL, max retries, and backoff factor
        self.ttl = ttl
        self.max_retries: int = max_retries
        self.retry_backoff_factor: float = retry_backoff_factor
        self.session = requests_retry_session(
            retries=self.max_retries, backoff_factor=self.retry_backoff_factor
        )

    @log_time(prefix=LOG_PREFIX)
    def _make_request(
        self,
        endpoint: str,
        method: str = "GET",
        payload: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Internal method to handle requests to the signed URL server."""
        url = urljoin(self.base_url, endpoint)
        try:
            response = self.session.request(
                method, url, headers=headers, json=payload, timeout=MAX_TIMEOUT
            )
            response.raise_for_status()
            return response.json()
        except RequestException as e:
            raise RuntimeError(f"Error during request to {url}: {e}") from e

    @log_time(prefix=LOG_PREFIX)
    def _make_server_api_call(
        self,
        endpoint: SignedURLServerEndpoint,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Get a signed URL for the specified operation and URI."""
        query_string = urlencode(params or {})
        endpoint_with_params = f"{endpoint.value}?{query_string}"
        return self._make_request(
            endpoint=endpoint_with_params, method="GET", payload=None, headers=headers
        )

    @log_time(prefix=LOG_PREFIX)
    def _upload_data(
        self,
        signed_url: str,
        data: Union[bytes, io.BufferedReader],
        headers: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Upload data to the specified storage path using a signed URL.

        Args:
            signed_url: str: The signed URL to upload the data to.
            data: Bytes or IO: The data to upload.
        """
        if isinstance(data, io.BufferedReader):
            if os.fstat(data.fileno()).st_size == 0:
                data = b""
        try:
            headers = headers or {}
            headers["Content-Type"] = "application/octet-stream"
            response = self.session.put(
                url=signed_url,
                headers=headers,
                data=data,
                timeout=REQUEST_TIMEOUT,
            )
            response.raise_for_status()
        except Exception as e:
            raise RuntimeError(f"Failed to upload data: {e}") from e

    @log_time(prefix=LOG_PREFIX)
    def upload_from_bytes(self, data: bytes, storage_uri: str) -> str:
        """Upload bytes to the specified storage path using a signed URL."""
        signed_object = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.WRITE,
            params={"uri": storage_uri, "expiryInSeconds": self.ttl},
            headers=self.signed_url_server_headers,
        )
        pre_signed_object_dto = SignedURLAPIResponseDto.parse_obj(signed_object)
        self._upload_data(
            signed_url=pre_signed_object_dto.signed_url,
            data=data,
            headers=pre_signed_object_dto.headers,
        )
        return storage_uri

    @log_time(prefix=LOG_PREFIX)
    def upload(self, file_path: str, storage_uri: str) -> str:
        """Upload a file to the specified storage path using a signed URL."""
        logger.info(f"{LOG_PREFIX} Uploading {file_path} to {storage_uri}")
        response = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.WRITE,
            params={"uri": storage_uri, "expiryInSeconds": self.ttl},
            headers=self.signed_url_server_headers,
        )
        pre_signed_object_dto = SignedURLAPIResponseDto.parse_obj(response)
        with open(file_path, "rb") as file:
            self._upload_data(
                signed_url=pre_signed_object_dto.signed_url,
                data=file,
                headers=pre_signed_object_dto.headers,
            )
        return storage_uri

    def create_multipart_upload(
        self, storage_uri: str, num_parts: int
    ) -> SignedURLMultipartUploadAPIResponseDto:
        response = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.CREATE_MUTLIPART_UPLOAD,
            params={
                "path": storage_uri,
                "numParts": num_parts,
                "partExpiryInSeconds": self.ttl,
                "finalizationExpiryInSeconds": MULTIPART_UPLOAD_FINALIZE_SIGNED_URL_TIMEOUT,
            },
            headers=self.signed_url_server_headers,
        )

        return SignedURLMultipartUploadAPIResponseDto.parse_obj(response)

    @log_time(prefix=LOG_PREFIX)
    def _download_file(
        self,
        signed_url: str,
        local_path: Optional[str] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Optional[bytes]:
        """Common method to download a file using a signed URL."""
        try:
            if headers is None:
                headers = {"Content-Type": "application/octet-stream"}
            else:
                headers["Content-Type"] = "application/octet-stream"
            response = self.session.get(
                signed_url,
                stream=True,
                headers=headers,
                timeout=REQUEST_TIMEOUT,
            )
            response.raise_for_status()
            if local_path:
                with open(local_path, "wb") as file:
                    for chunk in response.iter_content(chunk_size=8192):
                        file.write(chunk)
                return None
            return response.content
        except RequestException as e:
            raise RuntimeError(f"Failed to download file from {signed_url}: {e}") from e

    @log_time(prefix=LOG_PREFIX)
    def download(self, storage_uri: str, local_path: str) -> Optional[str]:
        """Download a file from the specified storage path to a local path using a signed URL."""
        logger.info(f"{LOG_PREFIX} Downloading {storage_uri} to {local_path}")
        response = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.READ,
            params={"uri": storage_uri, "expiryInSeconds": self.ttl},
            headers=self.signed_url_server_headers,
        )
        presigned_object = SignedURLAPIResponseDto.parse_obj(response)
        self._download_file(
            signed_url=presigned_object.signed_url,
            local_path=local_path,
            headers=presigned_object.headers,
        )
        return local_path

    @log_time(prefix=LOG_PREFIX)
    def download_to_bytes(self, storage_uri: str) -> Optional[bytes]:
        """Download a file from the specified storage path and return it as bytes."""
        response = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.READ,
            params={"uri": storage_uri, "expiryInSeconds": self.ttl},
            headers=self.signed_url_server_headers,
        )
        presigned_object = SignedURLAPIResponseDto.parse_obj(response)
        return self._download_file(
            signed_url=presigned_object.signed_url, headers=presigned_object.headers
        )

    @log_time(prefix=LOG_PREFIX)
    def exists(self, uri: str) -> bool:
        """Check if a file exists at the specified path."""
        response = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.EXISTS,
            params={"uri": uri},
            headers=self.signed_url_server_headers,
        )
        return SignedURLExistsAPIResponseDto.parse_obj(response).exists

    @log_time(prefix=LOG_PREFIX)
    def is_directory(self, uri: str) -> bool:
        """Check if the specified URI is a directory."""
        response = self._make_server_api_call(
            endpoint=SignedURLServerEndpoint.IS_DIRECTORY,
            params={"path": uri},
            headers=self.signed_url_server_headers,
        )
        is_directory = SignedURLIsDirectoryAPIResponseDto.parse_obj(
            response
        ).is_directory
        logger.info(f"{LOG_PREFIX} Path {uri} is a directory: {is_directory}")
        return is_directory

    @log_time(prefix=LOG_PREFIX)
    def list_files(
        self, path: str, detail: bool = False, max_results: int = 1000
    ) -> Union[List[FileInfo], List[str]]:
        """List files in the specified directory."""
        token = ""
        items: List[FileInfo] = []
        # Fetch all files in the specified path, in pages of max_results
        while True:
            response = self._make_server_api_call(
                endpoint=SignedURLServerEndpoint.LIST_FILES,
                params={"path": path, "maxResults": max_results, "pageToken": token},
                headers=self.signed_url_server_headers,
            )
            response_obj = PagedList.parse_obj(response)
            items.extend(response_obj.items)
            token = response_obj.token
            if not token:
                break

        # Return the items or paths based on the detail flag
        return items if detail else [item.path for item in items]



from __future__ import annotations

from typing import Any, Generator, Optional

import requests

from truefoundry.autodeploy.tools.base import (
    Event,
    Message,
    RequestEvent,
    ResponseEvent,
    Rule,
    Tool,
)
from truefoundry.pydantic_v1 import Field


class SendRequest(Tool):
    description = """
Send an HTTP request.
"""

    def __init__(self):
        self.call_count = 0

    class Request(RequestEvent):
        method: str
        url: str

    class Response(ResponseEvent):
        response_code: Optional[int] = Field(None, description="Response Code")
        response_body: Optional[str] = None
        error: Optional[str] = Field(None, description="Error.")

    def run(self, request: SendRequest.Request) -> Generator[Event, Any, ResponseEvent]:
        self.call_count += 1
        yield Rule(message="[bold green]Endpoint Verification[/]")
        yield Message(
            message=f"[bold cyan]Testing:[/] Sending a [magenta]{request.method.upper()}[/] request to [magenta]{request.url}[/]"
        )
        try:
            response = requests.request(request.method.lower(), url=request.url)
            yield Message(
                message=f"[bold green]Success:[/] Received response with status code [magenta]{response.status_code}[/]"
            )
            return SendRequest.Response(
                response_code=response.status_code,
                response_body=response.text[-50:],
            )
        except Exception as ex:
            yield Message(
                message="[red]Alert:[/] Request could not be completed successfully."
            )
            return SendRequest.Response(
                error=str(ex),
            )



import atexit
import threading
import weakref
from typing import TYPE_CHECKING, Dict, Optional

from truefoundry.common.request_utils import urllib3_retry
from truefoundry.common.session import Session
from truefoundry.common.utils import (
    get_tfy_servers_config,
    get_user_agent,
    relogin_error_message,
)
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ApiClient,
    Configuration,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.logger import logger

if TYPE_CHECKING:
    from truefoundry.ml.mlfoundry_run import MlFoundryRun

SESSION_LOCK = threading.RLock()
ACTIVE_SESSION: Optional["MLFoundrySession"] = None


class ActiveRuns:
    def __init__(self):
        self._active_runs: Dict[str, weakref.ReferenceType["MlFoundryRun"]] = {}

    def add_run(self, run: "MlFoundryRun"):
        with SESSION_LOCK:
            self._active_runs[run.run_id] = weakref.ref(run)

    def remove_run(self, run: "MlFoundryRun"):
        with SESSION_LOCK:
            if run.run_id in self._active_runs:
                del self._active_runs[run.run_id]

    def close_active_runs(self):
        with SESSION_LOCK:
            for run_ref in list(self._active_runs.values()):
                run = run_ref()
                if run and run.auto_end:
                    run.end()
            self._active_runs.clear()


ACTIVE_RUNS = ActiveRuns()
atexit.register(ACTIVE_RUNS.close_active_runs)


class MLFoundrySession(Session):
    def _assert_not_closed(self):
        if self._closed:
            raise MlFoundryException(
                "This session has been deactivated.\n"
                "At a time only one `client` (received from "
                "`truefoundry.ml.get_client()` function call) can be used"
            )

    def close(self):
        global ACTIVE_RUNS
        logger.debug("Closing existing session")
        ACTIVE_RUNS.close_active_runs()
        super().close()

    @classmethod
    def new(cls) -> "MLFoundrySession":
        global ACTIVE_SESSION
        with SESSION_LOCK:
            new_session = cls()
            if ACTIVE_SESSION and ACTIVE_SESSION == new_session:
                return ACTIVE_SESSION

            if ACTIVE_SESSION:
                ACTIVE_SESSION.close()

            ACTIVE_SESSION = new_session
            logger.info(
                "Logged in to %r as %r",
                new_session.tfy_host,
                new_session.user_info.user_id,
            )

            return ACTIVE_SESSION


class MLFoundryServerApiClient(ApiClient):
    def __init__(self, session: Optional[MLFoundrySession] = None, *args, **kwargs):
        self.session = session
        super().__init__(*args, **kwargs)

    @classmethod
    def from_session(cls, session: MLFoundrySession) -> "MLFoundryServerApiClient":
        mlfoundry_server_url = get_tfy_servers_config(
            session.tfy_host
        ).mlfoundry_server_url
        configuration = Configuration(
            host=mlfoundry_server_url.rstrip("/"),
            access_token=session.access_token,
        )
        configuration.retries = urllib3_retry(retries=2)
        api_client = cls(session=session, configuration=configuration)
        api_client.user_agent = get_user_agent()
        return api_client

    def _ensure_session(self):
        if self.session is None:
            raise MlFoundryException(
                relogin_error_message(
                    "No active session found. Perhaps you are not logged in?",
                )
            )

    @property
    def tfy_host(self) -> str:
        self._ensure_session()
        assert self.session is not None
        return self.session.tfy_host

    @property
    def access_token(self) -> str:
        self._ensure_session()
        assert self.session is not None
        return self.session.access_token


def _get_api_client(
    session: Optional[MLFoundrySession] = None,
    allow_anonymous: bool = False,
) -> MLFoundryServerApiClient:
    global ACTIVE_SESSION

    session = session or ACTIVE_SESSION
    if session is None:
        if allow_anonymous:
            return MLFoundryServerApiClient(session=None)
        else:
            raise MlFoundryException(
                relogin_error_message(
                    "No active session found. Perhaps you are not logged in?",
                )
            )
    return MLFoundryServerApiClient.from_session(session)



import difflib
import pprint
from typing import Iterator, List

from rich.console import Console, ConsoleOptions, RenderResult
from rich.text import Text


def rewrite_line(line, line_to_rewrite, prev_marker):
    marker_style_map = {
        "+": {
            " ": "green",
            "+": "white on green",
            "^": "white on green",
        },
        "-": {
            " ": "red",
            "-": "white on red",
            "^": "white on red",
        },
    }
    new_line = Text("")
    current_span = []
    # Differ lines start with a 2 letter code, so skip past that
    prev_char = line[2]
    for idx, char in enumerate(line[2:], start=2):
        if prev_marker in ("+", "-"):
            if char != prev_char:
                style = marker_style_map.get(prev_marker, {}).get(prev_char, None)
                if style is not None:
                    new_line.append_text(Text("".join(current_span), style=style))
                current_span = []
            if idx - 2 < len(line_to_rewrite):
                current_span.append(line_to_rewrite[idx - 2])
        prev_char = char

    # Lines starting with ? aren't guaranteed to be the same length as the lines before them
    #  so some characters may be left over. Add any leftover characters to the output.
    # subtract 2 for code at start
    remaining_index = idx - 2
    if prev_marker == "-":
        new_line.append_text(Text(line_to_rewrite[remaining_index:], style="red"))
    elif prev_marker == "+":
        new_line.append_text(Text(line_to_rewrite[remaining_index:], style="green"))
    return new_line


def build_symbolic_unified_diff(diff: List[str]) -> RenderResult:
    output_lines = []
    style = "grey"
    last_style = style
    for line in diff:
        if line.startswith("+ "):
            style = "green"
            output_line = f"+ {line[2:]}"
        elif line.startswith("- "):
            style = "red"
            output_line = f"- {line[2:]}"
        elif line.startswith("? "):
            if last_style == "red":
                output_line = line[:-1].replace("+", "-")
            elif last_style == "green":
                output_line = line[:-1].replace("-", "+")
        else:
            output_line = line
            style = "gray"
        output_lines.append(Text(output_line, style=style))
        last_style = style if style != "gray" else last_style
    return output_lines


def build_unified_diff(diff: List[str]) -> RenderResult:
    prev_marker = ""
    output_lines: List[Text] = []
    for line in diff:
        if line.startswith("- "):
            output_lines.append(Text(line[2:], style="red"))
        elif line.startswith("+ "):
            output_lines.append(Text(line[2:], style="green"))
        elif line.startswith("? "):
            line_to_rewrite = output_lines[-1].plain
            output_lines[-1] = rewrite_line(line, line_to_rewrite, prev_marker)
        else:
            output_lines.append(Text(line[2:], style="#949494"))
        prev_marker = line[0]
    return output_lines


def llm_unified_diff(diff: List[str]) -> RenderResult:
    prev_marker = ""
    output_lines: List[Text] = []
    for line in diff:
        if len(line) < 1:
            continue
        if line.startswith("---"):
            output_lines.append(Text(line, style="dim"))
        elif line.startswith("+++"):
            output_lines.append(Text(line, style="dim"))
        elif line.startswith("@@"):
            output_lines.append(Text(line, style="bright_blue"))
        elif line.startswith("-"):
            output_lines.append(Text(line[1:], style="red"))
        elif line.startswith("+"):
            output_lines.append(Text(line[1:], style="green"))
        elif line.startswith("?"):
            line_to_rewrite = output_lines[-1].plain
            output_lines[-1] = rewrite_line(line, line_to_rewrite, prev_marker)
        else:
            output_lines.append(Text(line[1:], style="#949494"))
        prev_marker = line[0]
    return output_lines


class LLMDiff:
    def __init__(self, llm_diff: str) -> None:
        self.diff = llm_diff.splitlines()

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        return llm_unified_diff(diff=self.diff)


class Diff:
    """Constructs a Diff object to render diff-highlighted code."""

    def __init__(
        self,
        lhs: object,
        rhs: object,
        width: int,
        show_symbols: bool = False,
    ) -> None:
        self.width = width
        self.lhs = lhs if isinstance(lhs, str) else pprint.pformat(lhs, width=width)
        self.rhs = rhs if isinstance(rhs, str) else pprint.pformat(rhs, width=width)
        self.show_symbols = show_symbols

    @property
    def sides_are_different(self) -> bool:
        return self.lhs != self.rhs

    def raw_unified_diff(self) -> Iterator[str]:
        differ = difflib.Differ()
        lines_lhs = self.lhs.splitlines()
        lines_rhs = self.rhs.splitlines()
        return differ.compare(lines_lhs, lines_rhs)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        diff = self.raw_unified_diff()
        if self.show_symbols:
            result = build_symbolic_unified_diff(diff)
        else:
            result = build_unified_diff(diff)
        yield from result



import warnings
from typing import Any, Literal, Union

from truefoundry.common.warnings import TrueFoundryDeprecationWarning
from truefoundry.deploy._autogen import models
from truefoundry.deploy.lib.model.entity import Deployment
from truefoundry.deploy.v2.lib.deploy import deploy_component
from truefoundry.deploy.v2.lib.patched_models import LocalSource
from truefoundry.pydantic_v1 import BaseModel, Field, conint, root_validator, validator

_TRIGGER_ON_DEPLOY_DEPRECATION_MESSAGE = """
Setting `trigger_on_deploy` in manifest has been deprecated and the field will be removed in future releases.

Please remove it from the spec and instead use

`trigger_on_deploy` argument on `.deploy`

E.g.

```
job = Job(...)  # remove `trigger_on_deploy` from initialization
job.deploy(..., trigger_on_deploy={arg_value})
```

OR

`{flag}` option on `tfy deploy`

E.g.

```
tfy deploy -f truefoundry.yaml {flag}
```
"""


def _warn_if_trigger_on_deploy_used(_klass, v: Any) -> Any:
    if v is not None:
        # v is the value of trigger_on_deploy, which is also the arg_value for the message
        flag = "--trigger-on-deploy" if v else "--no-trigger-on-deploy"
        warnings.warn(
            _TRIGGER_ON_DEPLOY_DEPRECATION_MESSAGE.format(arg_value=v, flag=flag),
            TrueFoundryDeprecationWarning,
            stacklevel=2,
        )
    return v


class DeployablePatchedModelBase(BaseModel):
    class Config:
        extra = "forbid"

    def deploy(
        self,
        workspace_fqn: str,
        wait: bool = True,
        force: bool = False,
        trigger_on_deploy: bool = False,
    ) -> Deployment:
        return deploy_component(
            component=self,
            workspace_fqn=workspace_fqn,
            wait=wait,
            force=force,
            trigger_on_deploy=trigger_on_deploy,
        )


class Service(models.Service, DeployablePatchedModelBase):
    type: Literal["service"] = "service"
    resources: models.Resources = Field(default_factory=models.Resources)
    # This is being patched because cue export marks this as a "number"
    replicas: Union[conint(ge=0, le=100), models.ServiceAutoscaling] = Field(  # type: ignore[valid-type]
        1,
        description="+label=Replicas\n+usage=Replicas of service you want to run\n+icon=fa-clone\n+sort=3",
    )


class Job(models.Job, DeployablePatchedModelBase):
    type: Literal["job"] = "job"
    resources: models.Resources = Field(default_factory=models.Resources)

    @validator("trigger_on_deploy")
    def _warn_if_trigger_on_deploy_used(cls, v: Any) -> Any:
        return _warn_if_trigger_on_deploy_used(cls, v)


class SparkJob(models.SparkJob, DeployablePatchedModelBase):
    type: Literal["spark-job"] = "spark-job"


class Notebook(models.Notebook, DeployablePatchedModelBase):
    type: Literal["notebook"] = "notebook"
    resources: models.Resources = Field(default_factory=models.Resources)


class Codeserver(models.Codeserver, DeployablePatchedModelBase):
    type: Literal["codeserver"] = "codeserver"
    resources: models.Resources = Field(default_factory=models.Resources)


class RStudio(models.RStudio, DeployablePatchedModelBase):
    type: Literal["rstudio"] = "rstudio"
    resources: models.Resources = Field(default_factory=models.Resources)


class Helm(models.Helm, DeployablePatchedModelBase):
    type: Literal["helm"] = "helm"


class Volume(models.Volume, DeployablePatchedModelBase):
    type: Literal["volume"] = "volume"


class ApplicationSet(models.ApplicationSet, DeployablePatchedModelBase):
    type: Literal["application-set"] = "application-set"


class AsyncService(models.AsyncService, DeployablePatchedModelBase):
    type: Literal["async-service"] = "async-service"
    replicas: Union[conint(ge=0, le=100), models.AsyncServiceAutoscaling] = 1  # type: ignore[valid-type]
    resources: models.Resources = Field(default_factory=models.Resources)


class SSHServer(models.SSHServer, DeployablePatchedModelBase):
    type: Literal["ssh-server"] = "ssh-server"
    resources: models.Resources = Field(default_factory=models.Resources)


class Workflow(models.Workflow, DeployablePatchedModelBase):
    type: Literal["workflow"] = "workflow"
    source: Union[models.RemoteSource, models.LocalSource] = Field(
        default_factory=lambda: LocalSource(local_build=False)
    )

    def deploy(
        self, workspace_fqn: str, wait: bool = True, force: bool = False
    ) -> Deployment:
        from truefoundry.deploy.v2.lib.deploy_workflow import deploy_workflow

        return deploy_workflow(
            workflow=self, workspace_fqn=workspace_fqn, wait=wait, force=force
        )


class Application(models.Application, DeployablePatchedModelBase):
    @root_validator(pre=True)
    def _validate_spec(cls, values: Any) -> Any:
        if isinstance(values, dict) and "__root__" in values:
            root = values["__root__"]
            if (
                isinstance(root, dict)
                and root.get("type") == "job"
                and root.get("trigger_on_deploy") is not None
            ):
                _warn_if_trigger_on_deploy_used(cls, root.get("trigger_on_deploy"))
        return values

    def deploy(
        self,
        workspace_fqn: str,
        wait: bool = True,
        force: bool = False,
        trigger_on_deploy: bool = False,
    ) -> Deployment:
        if isinstance(self.__root__, models.Workflow):
            from truefoundry.deploy.v2.lib.deploy_workflow import deploy_workflow

            return deploy_workflow(
                workflow=self.__root__,
                workspace_fqn=workspace_fqn,
                wait=wait,
                force=force,
            )
        else:
            return deploy_component(
                component=self.__root__,
                workspace_fqn=workspace_fqn,
                wait=wait,
                force=force,
                trigger_on_deploy=trigger_on_deploy,
            )
